# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import inspect

import memory_profiler as mp

from klio_exec.commands.utils import wrappers


class KMemoryLineProfiler(wrappers.KLineProfilerMixin, mp.LineProfiler):
    @classmethod
    def _wrap_per_element_func(cls, func, stream=None):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            prof = cls(backend="psutil")
            try:
                return prof(func)(*args, **kwargs)
            finally:
                mp.show_results(prof, stream=stream)

        return wrapper

    @classmethod
    def _wrap_per_element_gen(cls, func, stream=None):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            prof = cls(backend="psutil")
            try:
                yield from prof(func)(*args, **kwargs)
            finally:
                mp.show_results(prof, stream=stream)

        return wrapper

    @classmethod
    def wrap_per_element(cls, func, **kwargs):
        if inspect.isgeneratorfunction(func):
            return cls._wrap_per_element_gen(func, **kwargs)
        return cls._wrap_per_element_func(func, **kwargs)

    @staticmethod
    def _wrap_maximum_func(prof, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return prof(func)(*args, **kwargs)

        return wrapper

    @staticmethod
    def _wrap_maximum_gen(prof, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            yield from prof(func)(*args, **kwargs)

        return wrapper

    @classmethod
    def wrap_maximum(cls, prof, func, **_):
        if inspect.isgeneratorfunction(func):
            return cls._wrap_maximum_gen(prof, func)
        return cls._wrap_maximum_func(prof, func)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import traceback


def _get_relevant_frames(tb):
    """Remove up to and including pytest-related frames from tb.

    Args:
        list(traceback.FrameSummary): complete traceback of code
            executed by pytest
    Returns:
        list(traceback.FrameSummary) of traceback frames not including
            the invocation of pytest
    """
    last_index = 0
    for i, frame in enumerate(reversed(tb)):
        if "pytest" in str(frame):
            last_index = len(tb) - i
            break
    return tb[last_index:]


class BaseKlioAuditStep(object):
    """
    A base class to be subclassed when creating new audit steps for
    auditing Klio jobs.

    Subclasses have access to the following properties:
     - self.job_dir
     - self.klio_config

    All other inspection or modification of the Python environment
    should happen using global state, as pytest itself is run in between
    before_tests and after_tests.
    """

    AUDIT_STEP_NAME = None

    def __init__(self, job_dir, klio_config, tw):
        self.klio_config = klio_config
        self.job_dir = job_dir
        self.tw = tw
        self.errored = False
        self.warned = False

    @staticmethod
    def get_description():
        """
        User-facing description of audit step shown in `klio job audit --list`.

        If not implemented, the audit step class's docstring is used.
        If there is no class-level docstring, description defaults to
        "No description".
        """
        pass

    def _emit(self, message, tb=None, **kw):
        msg = "[{}]: {}\n".format(self.AUDIT_STEP_NAME, message)
        if tb:
            tb = _get_relevant_frames(tb)
            tb_fmtd = "\n".join(traceback.format_list(tb))
            msg = "{}{}\n".format(msg, tb_fmtd)

        self.tw.write(msg, **kw)

    def emit_warning(self, warning, tb=None, **kw):
        """Emit audit warning and set `self.warned` to True.

        Excludes any pytest related stacktraces in tb.

        Args:
              warning(str): Warning to emit. Prefixed to tb.
              tb(traceback object): Traceback to emit.
              kw(dict): Any attributes to pass onto `TerminalWriter.write`.
        """
        self.warned = True

        kwargs = {"yellow": True}
        kwargs.update(kw)
        self._emit(warning, tb=tb, **kwargs)

    def emit_error(self, error, tb=None, **kw):
        """Emit audit error and set `self.errored` to True.

        Excludes any pytest related stacktraces in tb.

        Args:
              warning(str): Warning to emit. Prefixed to tb.
              tb(traceback object): Traceback to emit.
              kw(dict): Any attributes to pass onto `TerminalWriter.write`.
        """
        self.errored = True

        kwargs = {"red": True}
        kwargs.update(kw)
        self._emit(error, tb=tb, **kwargs)

    def before_tests(self):
        """Any setup needed for test."""
        pass

    def after_tests(self):
        """Evaluate the results of auditing and emit results.

        Users can use the `emit_warning` and `emit_error` methods
        here to simplify reporting results. Implementation required.
        """
        raise NotImplementedError("Subclasses must implement after_tests!")



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect

import attr
import pkg_resources

from py import io


@attr.s
class KlioPlugin(object):
    plugin_name = attr.ib(type=str)
    description = attr.ib(type=str)
    package_name = attr.ib(type=str)
    package_version = attr.ib(type=str)
    module_path = attr.ib(type=str)


# TODO: in the future, add functionality to toggle & configure audit
#       steps in a job's klio-job.yaml file (@lynn)
def load_plugins_by_namespace(namespace):
    """Loads audit steps defined in `setup.py` under a given namespace.

    Args:
        namespace (str): namespace under which to look for plugins.
    Returns:
        Loaded plugin objects (list).
    """
    return [e.load() for e in pkg_resources.iter_entry_points(namespace)]


def _get_plugins_by_namespace(namespace):
    entrypoints = pkg_resources.iter_entry_points(namespace)
    for ep in entrypoints:
        # Need to actually load the plugin in order to get its location,
        # as well as class attributes, like name & description
        loaded = ep.load()

        desc = loaded.get_description() or loaded.__doc__
        if desc is None:
            desc = "No description."

        yield KlioPlugin(
            plugin_name=loaded.AUDIT_STEP_NAME,
            description=desc,
            package_name=ep.dist.project_name,
            package_version=ep.dist.parsed_version,
            module_path=inspect.getfile(loaded),
        )


def print_plugins(namespace, tw=None):
    plugin_meta = (
        " -- via {package_name} (v{package_version}) -- {module_path}\n"
    )
    plugin_desc = "\t{desc}\n\n"

    if not tw:
        tw = io.TerminalWriter()

    loaded_plugins = _get_plugins_by_namespace(namespace)
    for plugin in loaded_plugins:
        meta = plugin_meta.format(
            package_name=plugin.package_name,
            package_version=plugin.package_version,
            module_path=plugin.module_path,
        )
        tw.write(plugin.plugin_name, blue=True, bold=True)
        tw.write(meta, green=True)
        tw.write(plugin_desc.format(desc=plugin.description))



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import time

import pytest

from py import io

from klio_exec.commands.utils import plugin_utils


PLUGIN_NAMESPACE = "klio.plugins.audit"


def _run_pytest(tw):
    tw.write("Running tests for audit validation...\n")
    # Test-style verification
    exit_code = pytest.main(["-qq"])
    pytest_failed = exit_code != 0

    if pytest_failed:
        tw.write("PyTest failed!\n", yellow=True)
    else:
        tw.write("PyTest passed!\n")

    return pytest_failed


def _get_audit_steps(job_dir, config_obj, tw):
    loaded_steps = plugin_utils.load_plugins_by_namespace(PLUGIN_NAMESPACE)
    return [step(job_dir, config_obj, tw) for step in loaded_steps]


def list_audit_steps(ctx, param, value):
    # this is a click.option callback where it automatically passes in
    # context, param name, and param value
    if not value or ctx.resilient_parsing:
        return
    tw = io.TerminalWriter()
    tw.hasmarkup = True
    tw.sep("=", "Installed audit steps")
    plugin_utils.print_plugins(PLUGIN_NAMESPACE, tw)
    ctx.exit()


def audit(job_dir, config_obj):
    tw = io.TerminalWriter()
    tw.hasmarkup = True
    tw.write("Auditing your Klio job...\n")

    try:
        audit_steps = _get_audit_steps(job_dir, config_obj, tw)
        for step in audit_steps:
            step.before_tests()

        pytest_failed = _run_pytest(tw)

        tw.sep("=", "audit session starts", bold=True)

        start = time.time()

        # TODO: Currently, we're running all of the audit steps in
        # the same Python process and there's no isolation. Do we want
        # to run the tests once per step instead?
        for step in audit_steps:
            step.after_tests()

            if not any([step.errored, step.warned]):
                msg = "[{}]: audit step passed!\n".format(step.AUDIT_STEP_NAME)
                tw.write(msg, green=True)

        end = time.time()

        failed_steps = [step for step in audit_steps if step.errored]
        warned_steps = [step for step in audit_steps if step.warned]

        error_count = len(failed_steps)
        warning_count = len(warned_steps)

        finished_msg = "{0} errors, {1} warnings in {2:.3f} seconds".format(
            error_count, warning_count, end - start
        )
        color = "green"
        if error_count:
            color = "red"
        elif warning_count:
            color = "yellow"
        kwargs = {
            color: True,
            "bold": True,
        }
        tw.sep("=", finished_msg, **kwargs)

    except Exception:
        # use logging instead of tw to easily get traceback info
        logging.error(
            "Unable to run the audit command due to:\n", exc_info=True
        )

        raise SystemExit(1)

    if not pytest_failed and not error_count:
        if not warning_count:
            tw.write("Good job! Your job does not exhibit any known issues.\n")
        else:
            tw.write(
                "Cool! Your job has warnings, but no errors. "
                "Please check the warnings.\n",
                yellow=True,
            )
    else:
        tw.sep(
            "You have errors in your job. Please fix and try again.\n",
            red=True,
        )
        raise SystemExit(1)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections
import logging
import os

import click
import yaml

# WARNING: be careful with what is imported here.  Any decorators that read
# config cannot be imported here (even transitively) since they will attempt to
# read the config object before it exists.
from klio.transforms import core as klio_transforms_core
from klio_core import config
from klio_core import options as core_options
from klio_core import utils as core_utils

from klio_exec import options
from klio_exec.commands import audit
from klio_exec.commands import stop


# TODO remove 'INFO:root:' bit from cli status logs.
# TODO make 'ERROR' logs red colored?
logging.getLogger().setLevel(logging.INFO)


RuntimeConfig = collections.namedtuple(
    "RuntimeConfig", ["image_tag", "direct_runner", "update", "blocking"]
)


@click.group()
def main():
    pass  # pragma: no cover


def _get_config(config_path):
    try:
        with open(config_path) as f:
            return yaml.safe_load(f)
    except IOError as e:
        logging.error(e)
        raise SystemExit(1)


# TODO: remove when after internal calls are removed
def _compare_runtime_to_buildtime_config(klio_config):
    pass


@main.command("run")
@core_options.image_tag
@core_options.direct_runner
@options.blocking
@core_options.update
@core_utils.with_klio_config
def run_pipeline(
    image_tag, direct_runner, update, klio_config, config_meta, blocking
):

    # RunConfig ensures config is pickled and sent to worker.  Note this
    # depends on save_main_session being True
    # Notice, this is currently unused due to dataflow pickling issues, leaving
    # in for compatibility until a fix/alternative solution is in place
    klio_transforms_core.RunConfig.set(klio_config)

    # This can only be imported after RunConfig is set since it will end up
    # importing classes that may (or do) attempt to read it
    from klio_exec.commands import run

    if update is None:  # if it's not explicitly set in CLI, look at config
        update = klio_config.pipeline_options.update
    if blocking is None:  # if it's not explicitly set in CLI, look at config
        blocking = klio_config.job_config.blocking

    runtime_conf = RuntimeConfig(image_tag, direct_runner, update, blocking)

    klio_pipeline = run.KlioPipeline(
        klio_config.job_name, klio_config, runtime_conf
    )
    klio_pipeline.run()


@main.command("stop")
@options.config_file
def stop_job(config_file):
    job_dir = os.path.abspath(os.getcwd())
    config_file = config_file or "klio-job.yaml"
    config_path = os.path.join(job_dir, config_file)
    config_data = _get_config(config_path)
    conf_obj = config.KlioConfig(config_data)
    # TODO: make this a click option once draining is supported @lynn
    strategy = "cancel"
    stop.stop(conf_obj, strategy)


@main.command("test", context_settings=dict(ignore_unknown_options=True))
@options.config_file
@click.argument("pytest_args", nargs=-1, type=click.UNPROCESSED)
def test_job(config_file, pytest_args):
    """Thin wrapper around pytest. Any arguments after -- are passed through.
    """
    import os
    import pytest

    config_path = config_file or "klio-job.yaml"
    config_data = _get_config(config_path)
    conf_obj = config.KlioConfig(config_data)

    # RunConfig ensures config is pickled and sent to worker.  Note this
    # depends on save_main_session being True
    klio_transforms_core.RunConfig.set(conf_obj)

    # NOTE: we assume that test_job is the only method called in this
    # subprocess, so setting KLIO_TEST_MODE will not impact subsequent
    # method calls
    os.environ["KLIO_TEST_MODE"] = "true"
    exit_code = pytest.main(list(pytest_args))

    if exit_code != 0:
        raise SystemExit("Tests failed with exit code %s" % exit_code)


def _require_profile_input_data(input_file, entity_ids):
    # Note: can't use something like MutuallyExclusiveOption since entity IDs
    #       are click arguments, not options
    if not any([input_file, entity_ids]):
        msg = "Must provide `--input-file` or a list of entity IDs."
        raise click.UsageError(msg)

    if all([input_file, entity_ids]):
        msg = (
            "Illegal usage: `--input-file` is mutually exclusive with "
            "entity ID arguments."
        )
        raise click.UsageError(msg)


@main.group(
    "profile",
    help=(
        "Profile a job. NOTE: Requires klio-exec[debug] installed in the "
        "job's Docker image."
    ),
)
def profile_job():
    pass


# hidden command to only be invoked via a subprocess from
# `klioexec profile memory` or `klioexec profile cpu`
@profile_job.command("run-pipeline", hidden=True)
@options.input_file
@core_options.show_logs
@options.config_file
@click.argument("entity_ids", nargs=-1, required=False)
def _run_pipeline(input_file, show_logs, entity_ids, config_file):

    from klio_exec.commands import profile

    config_path = config_file or "klio-job.yaml"
    config_data = _get_config(config_path)
    klio_config = config.KlioConfig(config_data)

    # safety check, even though it should be invoked by another klioexec
    # command
    _require_profile_input_data(input_file, entity_ids)

    if not show_logs:
        logging.disable(logging.CRITICAL)

    klio_pipeline = profile.KlioPipeline(
        klio_config=klio_config, input_file=input_file, entity_ids=entity_ids
    )
    klio_pipeline.profile(what="run")


@profile_job.command(
    "memory",
    short_help="Profile overall memory usage.",
    help=(
        "Profile overall memory usage on an interval while running all "
        "Klio-based transforms."
    ),
)
@core_options.interval
@core_options.include_children
@core_options.multiprocess
@core_options.plot_graph
@options.input_file
@options.output_file
@core_options.show_logs
@options.config_file
@click.argument("entity_ids", nargs=-1, required=False)
def profile_memory(
    interval,
    include_children,
    multiprocess,
    input_file,
    output_file,
    plot_graph,
    show_logs,
    entity_ids,
    config_file,
):

    from klio_exec.commands import profile

    config_path = config_file or "klio-job.yaml"
    config_data = _get_config(config_path)
    klio_config = config.KlioConfig(config_data)

    _require_profile_input_data(input_file, entity_ids)

    klio_pipeline = profile.KlioPipeline(
        klio_config=klio_config,
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids,
    )
    kwargs = {
        "include_children": include_children,
        "multiprocess": multiprocess,
        "interval": interval,
        "show_logs": show_logs,
        "plot_graph": plot_graph,
    }
    output_png = klio_pipeline.profile(what="memory", **kwargs)
    if output_png:
        click.echo("Memory plot graph generated at: {}".format(output_png))


@profile_job.command(
    "memory-per-line",
    short_help="Profile memory usage per line.",
    help=(
        "Profile memory per line for every Klio-based transforms' process "
        "method."
    ),
)
@core_options.maximum
@core_options.per_element(show_default="True")
@options.input_file
@options.output_file
@core_options.show_logs
@options.config_file
@click.argument("entity_ids", nargs=-1, required=False)
def profile_memory_per_line(
    get_maximum,
    per_element,
    input_file,
    output_file,
    show_logs,
    entity_ids,
    config_file,
):
    from klio_exec.commands import profile

    config_path = config_file or "klio-job.yaml"
    config_data = _get_config(config_path)
    klio_config = config.KlioConfig(config_data)

    _require_profile_input_data(input_file, entity_ids)

    if not show_logs:
        logging.disable(logging.CRITICAL)

    klio_pipeline = profile.KlioPipeline(
        klio_config=klio_config,
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids,
    )
    klio_pipeline.profile(what="memory_per_line", get_maximum=get_maximum)


@profile_job.command(
    "cpu",
    short_help="Profile overall CPU usage.",
    help=(
        "Profile overall CPU usage on an interval while running all "
        "Klio-based transforms."
    ),
)
@core_options.interval
@options.input_file
@options.output_file
@core_options.plot_graph
@core_options.show_logs
@options.config_file
@click.argument("entity_ids", nargs=-1, required=False)
def profile_cpu(
    interval,
    input_file,
    output_file,
    plot_graph,
    show_logs,
    entity_ids,
    config_file,
):
    from klio_exec.commands import profile

    config_path = config_file or "klio-job.yaml"
    config_data = _get_config(config_path)
    klio_config = config.KlioConfig(config_data)

    _require_profile_input_data(input_file, entity_ids)

    klio_pipeline = profile.KlioPipeline(
        klio_config=klio_config,
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids,
    )
    kwargs = {
        "interval": interval,
        "show_logs": show_logs,
        "plot_graph": plot_graph,
    }
    output_png = klio_pipeline.profile(what="cpu", **kwargs)
    if output_png:
        click.echo("CPU plot graph generated at: {}".format(output_png))


@profile_job.command(
    "timeit",
    short_help="Profile wall time per line.",
    help=(
        "Profile wall time by every line for every Klio-based transforms' "
        "process method. NOTE: this uses the `line_profiler` package, not "
        "Python's `timeit` module."
    ),
)
@options.input_file
@options.output_file
@core_options.iterations
@core_options.show_logs
@options.config_file
@click.argument("entity_ids", nargs=-1, required=False)
def profile_wall_time(
    input_file, output_file, iterations, show_logs, entity_ids, config_file
):
    from klio_exec.commands import profile

    config_path = config_file or "klio-job.yaml"
    config_data = _get_config(config_path)
    klio_config = config.KlioConfig(config_data)

    _require_profile_input_data(input_file, entity_ids)

    if not show_logs:
        logging.disable(logging.CRITICAL)

    klio_pipeline = profile.KlioPipeline(
        klio_config=klio_config,
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids,
    )
    klio_pipeline.profile(what="timeit", iterations=iterations)


@main.command("audit", context_settings=dict(ignore_unknown_options=True))
@options.config_file
@click.option(
    "--list",
    is_flag=True,
    is_eager=True,
    expose_value=False,  # don't need to pass in parameter to audit_job func
    callback=audit.list_audit_steps,
    help="List available audit steps (does not run any audits).",
)
def audit_job(config_file):
    # NOTE: we assume that audit_job is the only method called in this
    # subprocess, so setting KLIO_TEST_MODE will not impact subsequent
    # method calls
    os.environ["KLIO_TEST_MODE"] = "true"

    job_dir = os.path.abspath(os.getcwd())
    config_data = _get_config(config_file or "klio-job.yaml")
    conf_obj = config.KlioConfig(config_data)
    audit.audit(job_dir, conf_obj)


if __name__ == "__main__":
    main()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

from klio_exec.commands.audit_steps import base


class MultithreadedTFUsage(base.BaseKlioAuditStep):
    """Use caution when running tensorflow in a multithreaded environment."""

    AUDIT_STEP_NAME = "multithreaded_tf"

    @property
    def _is_tensorflow_loaded(self):
        return any(["tensorflow" in module for module in sys.modules])

    @property
    def _is_job_single_threaded_per_container(self):
        exps = self.klio_config.pipeline_options.experiments
        return "worker_threads=1" in exps

    def after_tests(self):
        if not self._is_job_single_threaded_per_container:
            if self._is_tensorflow_loaded:
                self.emit_warning(
                    "TensorFlow usage detected within job, but "
                    "`worker_threads` is not explicitly set to 1 under "
                    "`pipeline_options.experiments` in the job's configuration "
                    "file! This can cause threading issues. Be careful."
                )


_init = MultithreadedTFUsage



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.audit_steps import numpy_broken_blas


@pytest.mark.parametrize(
    "worker_threads,platform,has_numpy,version,exp_emit_calls",
    (
        (1, None, False, None, 0),
        (2, "Darwin", False, None, 0),
        (0, "Linux", False, None, 0),
        (2, "Linux", False, None, 0),
        (2, "Linux", True, "1.17.0", 0),
        (2, "Linux", True, "1.16.3", 0),
        (2, "Linux", True, "1.16.2", 1),
    ),
)
def test_numpy_broken_blas_usage(
    worker_threads,
    platform,
    has_numpy,
    version,
    exp_emit_calls,
    klio_config,
    mock_emit_error,
    mocker,
    monkeypatch,
):
    if worker_threads:
        klio_config.pipeline_options.experiments = [
            "worker_threads={}".format(worker_threads)
        ]

    monkeypatch.setattr(numpy_broken_blas.platform, "system", lambda: platform)
    if has_numpy is False:
        mocker.patch.dict("sys.modules", {"numpy": None})
    else:
        monkeypatch.setattr("numpy.version.short_version", version)

    numpy_broken_blas_usage = numpy_broken_blas.NumPyBrokenBLASUsage(
        "job/dir", klio_config, "term_writer"
    )

    numpy_broken_blas_usage.after_tests()

    # don't care about the actual message
    assert exp_emit_calls == mock_emit_error.call_count



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import inspect


def _get_transform_error_msg(txf=None, entity_id=None, err_msg=None):
    # This error message is printed instead of logged since user may not
    # run with logs turned on
    return (
        "WARN: Error caught while profiling {txf}.process for "
        "entity ID {entity_id}: {err_msg}".format(
            txf=txf, entity_id=entity_id, err_msg=err_msg
        )
    )


def _print_user_exceptions_generator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        transform_name = args[0].__class__.__name__
        entity_id = args[1]
        result = None
        try:
            result = yield from func(*args, **kwargs)

        except Exception as e:
            msg = _get_transform_error_msg(
                txf=transform_name, entity_id=entity_id, err_msg=e
            )
            print(msg)

        return result

    return wrapper


def _print_user_exceptions_func(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        transform_name = args[0].__class__.__name__
        entity_id = args[1]
        result = None
        try:
            result = func(*args, **kwargs)

        except Exception as e:
            msg = _get_transform_error_msg(
                txf=transform_name, entity_id=entity_id, err_msg=e
            )
            print(msg)

        return result

    return wrapper


def print_user_exceptions(transforms):
    # Don't crap out if the process method errors; just continue profiling
    for txf in transforms:
        process_method = getattr(txf, "process")
        if inspect.isgeneratorfunction(process_method):
            process_method = _print_user_exceptions_generator(process_method)
        else:
            process_method = _print_user_exceptions_func(process_method)
        setattr(txf, "process", process_method)
        yield txf


# adapted from line_profiler; memory_profiler doesn't handle generator
# functions for some reason.
class KLineProfilerMixin(object):
    """Mixin for CPU & Memory line profilers."""

    def __call__(self, func):
        # Overwrite to handle generators in the same fashion as funcs
        self.add_function(func)
        if inspect.isgeneratorfunction(func):
            return self.wrap_generator(func)
        return self.wrap_function(func)

    def wrap_function(self, func):
        """Wrap a function to profile it."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.enable_by_count()
            try:
                return func(*args, **kwargs)
            finally:
                self.disable_by_count()

        return wrapper

    def wrap_generator(self, func):
        """Wrap a generator to profile it."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.enable_by_count()
            try:
                yield from func(*args, **kwargs)
            finally:
                self.disable_by_count()

        return wrapper



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from googleapiclient import errors as gerrors

from klio_exec.commands import stop


@pytest.fixture
def mock_discovery_client(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(stop.discovery, "build", lambda x, y: mock)
    return mock


@pytest.fixture
def config(mocker):
    pipeline_options = mocker.Mock(
        project="test-project", region="europe-west1"
    )
    return mocker.Mock(job_name="test-job", pipeline_options=pipeline_options)


@pytest.fixture
def jobs_response():
    return {
        "jobs": [
            {"name": "not-the-test-job"},
            {
                "id": "1234",
                "name": "test-job",
                "projectId": "test-project",
                "location": "europe-west1",
            },
        ]
    }


@pytest.fixture
def job(jobs_response):
    return jobs_response["jobs"][1]


@pytest.fixture
def mock_sleep(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(stop.time, "sleep", mock)
    return mock


@pytest.mark.parametrize("api_version", (None, "v1b3", "v2"))
def test_set_dataflow_client(mock_discovery_client, api_version):
    assert stop._client is None

    stop._set_dataflow_client(api_version)

    assert stop._client is not None
    assert mock_discovery_client == stop._client

    # cleanup
    setattr(stop, "_client", None)


# return the desired job, no jobs at all, or no jobs matching job name
@pytest.mark.parametrize("returns_jobs", (True, False, None))
def test_check_job_running(
    mock_discovery_client, returns_jobs, jobs_response, config, monkeypatch
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.list

    req.return_value.execute.return_value = {}
    if returns_jobs:
        req.return_value.execute.return_value = jobs_response
    elif returns_jobs is False:
        req.return_value.execute.return_value = {
            "jobs": [{"name": "not-the-test-job"}]
        }

    ret = stop._check_job_running(config)

    if returns_jobs:
        assert jobs_response["jobs"][1] == ret
    else:
        assert ret is None

    req.assert_called_once_with(
        projectId="test-project", location="europe-west1", filter="ACTIVE"
    )
    req.return_value.execute.assert_called_once_with()


def test_check_job_running_errors(
    mock_discovery_client, config, monkeypatch, caplog
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.list
    req.return_value.execute.side_effect = Exception("foo")

    stop._check_job_running(config)

    req.assert_called_once_with(
        projectId="test-project", location="europe-west1", filter="ACTIVE"
    )
    req.return_value.execute.assert_called_once_with()
    assert 2 == len(caplog.records)


@pytest.mark.parametrize(
    "state,pyver", (("drain", None), ("cancel", None), (None, 2), (None, 3))
)
def test_update_job_state(
    state, pyver, mock_discovery_client, job, monkeypatch
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    exp_state = state
    if not state:
        if pyver == 2:
            exp_state = "JOB_STATE_DRAINED"
        else:
            exp_state = "JOB_STATE_CANCELLED"
        monkeypatch.setitem(stop.JOB_STATE_MAP, "default", exp_state)

    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.update
    req.return_value.execute.return_value = None

    stop._update_job_state(job, state)

    job["requestedState"] = exp_state
    req.assert_called_once_with(
        jobId="1234",
        projectId="test-project",
        location="europe-west1",
        body=job,
    )
    req.return_value.execute.assert_called_once_with()


def test_update_job_state_400_error(
    mock_discovery_client, job, mock_sleep, mocker, monkeypatch, caplog
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.update

    mock_resp = mocker.Mock(status=400)
    req.return_value.execute.side_effect = gerrors.HttpError(mock_resp, b"foo")

    with pytest.raises(SystemExit):
        stop._update_job_state(job, "drain")

    assert 1 == req.return_value.execute.call_count
    assert 1 == len(caplog.records)
    assert not mock_sleep.call_count


def test_update_job_state_500_error(
    mock_discovery_client, job, mock_sleep, mocker, monkeypatch, caplog
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.update

    mock_resp = mocker.Mock(status=500)
    req.return_value.execute.side_effect = gerrors.HttpError(mock_resp, b"foo")

    with pytest.raises(SystemExit):
        stop._update_job_state(job, "drain")

    assert 4 == req.return_value.execute.call_count
    assert 4 == len(caplog.records)
    assert 3 == mock_sleep.call_count


def test_update_job_state_error(
    mock_discovery_client, job, mock_sleep, monkeypatch, caplog
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.update
    req.return_value.execute.side_effect = Exception("foo")

    with pytest.raises(SystemExit):
        stop._update_job_state(job, "cancel")

    assert 4 == req.return_value.execute.call_count
    assert 4 == len(caplog.records)
    assert 3 == mock_sleep.call_count


@pytest.mark.parametrize(
    "exec_side_effect",
    (
        (
            {"currentState": "JOB_STATE_CANCELLING"},
            {"currentState": "JOB_STATE_CANCELLED"},
        ),
        (Exception("foo"), {"currentState": "JOB_STATE_CANCELLED"}),
    ),
)
def test_watch_job_state(
    mock_discovery_client,
    mock_sleep,
    monkeypatch,
    caplog,
    job,
    exec_side_effect,
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)
    _projects_req = mock_discovery_client.projects.return_value
    req = _projects_req.locations.return_value.jobs.return_value.get
    req.return_value.execute.side_effect = exec_side_effect

    stop._watch_job_state(job)

    assert 2 == req.return_value.execute.call_count
    mock_sleep.assert_called_once_with(5)
    assert 1 == len(caplog.records)


def test_watch_job_state_raises(
    mock_discovery_client, monkeypatch, caplog, job
):
    monkeypatch.setattr(stop, "_client", mock_discovery_client)

    with pytest.raises(SystemExit):
        stop._watch_job_state(job, timeout=0)

    assert 1 == len(caplog.records)


@pytest.mark.parametrize("has_running_job", (True, False))
def test_stop(has_running_job, config, mocker, monkeypatch, job):
    mock_set_dataflow_client = mocker.Mock()
    monkeypatch.setattr(stop, "_set_dataflow_client", mock_set_dataflow_client)

    ret_val = None
    if has_running_job:
        ret_val = job
    mock_check_job_running = mocker.Mock(return_value=ret_val)
    monkeypatch.setattr(stop, "_check_job_running", mock_check_job_running)
    mock_update_job_state = mocker.Mock()
    monkeypatch.setattr(stop, "_update_job_state", mock_update_job_state)
    mock_watch_job_state = mocker.Mock()
    monkeypatch.setattr(stop, "_watch_job_state", mock_watch_job_state)

    stop.stop(config, "cancel")

    mock_set_dataflow_client.assert_called_once_with()
    mock_check_job_running.assert_called_once_with(config)
    if has_running_job:
        mock_update_job_state.assert_called_once_with(job, req_state="cancel")
        mock_watch_job_state.assert_called_once_with(job)
    else:
        mock_update_job_state.assert_not_called()
        mock_watch_job_state.assert_not_called()



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
from unittest import mock

import click
import pytest
import yaml

from click import testing

from klio_core import _testing as core_testing
from klio_core import config as kconfig


@pytest.fixture
def patch_os_getcwd(monkeypatch, tmpdir):
    test_dir = str(tmpdir.mkdir("testing"))
    monkeypatch.setattr(os, "getcwd", lambda: test_dir)
    return test_dir


@pytest.fixture
def mock_klio_config(mocker, monkeypatch, patch_os_getcwd):
    mock = core_testing.MockKlioConfig(
        cli, mocker, monkeypatch, patch_os_getcwd
    )
    mock.setup(_config(), "klio-job.yaml")
    return mock


@pytest.fixture
def mock_os_getcwd(monkeypatch):
    test_dir = "/test/dir"
    monkeypatch.setattr(os, "getcwd", lambda: test_dir)
    return test_dir


@pytest.fixture
def cli_runner():
    return testing.CliRunner()


def _config():
    return {
        "job_name": "klio-job-name",
        "job_config": {
            "inputs": [
                {
                    "data_location": "gs://test-in-data",
                    "topic": "test-in-topic",
                    "subscription": "test-sub",
                }
            ],
            "outputs": [
                {
                    "data_location": "gs://test-out-data",
                    "topic": "test-out-topic",
                }
            ],
        },
        "pipeline_options": {
            "streaming": True,
            "update": False,
            "worker_harness_container_image": "a-worker-image",
            "experiments": ["beam_fn_api"],
            "project": "test-gcp-project",
            "zone": "europe-west1-c",
            "region": "europe-west1",
            "staging_location": "gs://test-gcp-project-dataflow-tmp/staging",
            "temp_location": "gs://test-gcp-project-dataflow-tmp/temp",
            "max_num_workers": 2,
            "disk_size_gb": 32,
            "worker_machine_type": "n1-standard-2",
            "subnetwork": "some/path",
            "runner": "DataflowRunner",
        },
    }


@pytest.fixture
def config():
    return _config()


@pytest.fixture
def klio_config(config):
    return kconfig.KlioConfig(config)


# NOTE: Python decorators are evaluated on import, and so importing
# `klio_exec.commands.profile` (which imports `klio.transforms.helpers`, which
# imports `klio.transforms.decorators`) and `klio_exec.cli` triggers the  code
# in those decorators to get evaluated. Therefore, we must patch this part in
# order to import it, otherwise it will try to load the non-existant
# `/usr/src/config/.effective-klio-job.yaml`
mock_config = kconfig.KlioConfig(_config())
patcher = mock.patch(
    "klio.transforms.core.RunConfig.get", lambda: mock_config,
)
patcher.start()

from klio_exec.commands import profile  # noqa E402
from klio_exec import cli  # noqa E402


@pytest.fixture
def patch_get_config(monkeypatch, config):
    monkeypatch.setattr(cli, "_get_config", lambda x: config)


@pytest.fixture
def patch_klio_config(monkeypatch, klio_config):
    monkeypatch.setattr(cli.config, "KlioConfig", lambda x: klio_config)


@pytest.fixture
def patch_run_basic_pipeline(mocker, monkeypatch):
    mock = mocker.Mock()
    mocker.patch("klio_exec.commands.run.KlioPipeline.run", mock)
    return mock


@pytest.fixture
def mock_compare_runtime_to_buildtime_config(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(cli, "_compare_runtime_to_buildtime_config", mock)
    return mock


def _create_tmp_config_file(
    temp_dir_str: str, config_file: str, config
) -> str:
    """
    Creates a temporary config file

    Args:
        temp_dir_str: Temporary directory where to create file
        config_file: config file name
        config: config to write

    Returns:
        Temp config file path
    """
    # create a tmp file else click will complain it doesn't exist
    tmp_config_file = os.path.join(temp_dir_str, config_file)
    with open(tmp_config_file, "w") as f:
        yaml.dump(config, f)
    return tmp_config_file


def test_get_config(tmpdir, config):
    tmp_config = tmpdir.mkdir("klio-exec-testing").join("klio-job.yaml")
    tmp_config.write(yaml.dump(config))

    ret_config = cli._get_config(str(tmp_config))
    assert config == ret_config


def test_get_config_loaded_safely(tmpdir, config, mocker, monkeypatch):
    mock_safe_load = mocker.Mock()
    mock_safe_load.return_value = config
    monkeypatch.setattr(cli.yaml, "safe_load", mock_safe_load)
    tmp_config = tmpdir.mkdir("klio-exec-testing").join("klio-job.yaml")

    open_name = "klio_exec.cli.open"
    config_str = yaml.dump(config)
    m_open = mocker.mock_open(read_data=config_str)
    m = mocker.patch(open_name, m_open)

    cli._get_config(tmp_config.strpath)
    m.assert_called_once_with(tmp_config.strpath)
    mock_safe_load.assert_called_once_with(m_open.return_value)


def test_get_config_raises(tmpdir, caplog):
    tmp_config = tmpdir.mkdir("klio-exec-testing")

    does_not_exist = os.path.join(str(tmp_config), "klio-config.yaml")
    with pytest.raises(SystemExit):
        cli._get_config(does_not_exist)

    assert 1 == len(caplog.records)


@pytest.mark.parametrize("blocking", (True, False, None))
@pytest.mark.parametrize(
    "image_tag,direct_runner,update",
    (
        (None, True, False),
        (None, False, True),
        (None, False, None),
        ("a-tag", True, False),
        ("a-tag", False, False),
        ("a-tag", True, True),  # irrelevant, but CYA
    ),
)
def test_run_pipeline(
    image_tag,
    direct_runner,
    update,
    blocking,
    cli_runner,
    mock_klio_config,
    patch_run_basic_pipeline,
):
    mock_compare_runtime_to_buildtime_config.return_value = True
    runtime_conf = cli.RuntimeConfig(
        image_tag=None, direct_runner=False, update=None, blocking=None
    )
    cli_inputs = []
    if image_tag:
        cli_inputs.extend(["--image-tag", image_tag])
        runtime_conf = runtime_conf._replace(image_tag=image_tag)
    if direct_runner:
        cli_inputs.append("--direct-runner")
        runtime_conf = runtime_conf._replace(direct_runner=True)
    if update:
        cli_inputs.append("--update")
        runtime_conf = runtime_conf._replace(update=True)
    if update is False:
        cli_inputs.append("--no-update")
    if not update:  # if none or false
        runtime_conf = runtime_conf._replace(update=False)
    if blocking:
        cli_inputs.append("--blocking")
        runtime_conf._replace(blocking=True)
    if blocking is False:
        cli_inputs.append("--no-blocking")
    if not blocking:
        runtime_conf._replace(blocking=False)

    result = cli_runner.invoke(cli.run_pipeline, cli_inputs)
    core_testing.assert_execution_success(result)
    mock_klio_config.assert_calls()

    patch_run_basic_pipeline.assert_called_once_with()


@pytest.mark.parametrize(
    "config_file_override", (None, "klio-job2.yaml"),
)
def test_run_pipeline_conf_override(
    config_file_override,
    cli_runner,
    config,
    mock_klio_config,
    patch_run_basic_pipeline,
    caplog,
    tmpdir,
    monkeypatch,
):
    cli_inputs = []

    temp_dir = tmpdir.mkdir("testing123")
    temp_dir_str = str(temp_dir)
    monkeypatch.setattr(os, "getcwd", lambda: temp_dir_str)

    if config_file_override:
        # create a tmp file else click will complain it doesn't exist
        exp_conf_file = _create_tmp_config_file(
            temp_dir_str, config_file_override, config
        )
        cli_inputs.extend(["--config-file", exp_conf_file])
        mock_klio_config.setup(_config(), "klio-job.yaml", exp_conf_file)
    else:
        mock_klio_config.setup(_config(), "klio-job.yaml")

    result = cli_runner.invoke(cli.run_pipeline, cli_inputs)
    core_testing.assert_execution_success(result)
    mock_klio_config.assert_calls()

    patch_run_basic_pipeline.assert_called_once_with()

    assert 0 == len(caplog.records)


@pytest.mark.parametrize("config_file_override", (None, "klio-job2.yaml"))
def test_stop_job(
    config_file_override,
    mocker,
    monkeypatch,
    config,
    klio_config,
    patch_klio_config,
    cli_runner,
    tmpdir,
):
    mock_stop = mocker.Mock()
    monkeypatch.setattr(cli.stop, "stop", mock_stop)

    mock_get_config = mocker.Mock()
    monkeypatch.setattr(cli, "_get_config", mock_get_config)

    cli_inputs = []

    temp_dir = tmpdir.mkdir("testing")
    temp_dir_str = str(temp_dir)
    monkeypatch.setattr(os, "getcwd", lambda: temp_dir_str)

    exp_file = os.path.join(temp_dir_str, "klio-job.yaml")
    if config_file_override:
        # create a tmp file else click will complain it doesn't exist
        exp_file = _create_tmp_config_file(
            temp_dir_str, config_file_override, config
        )
        cli_inputs.extend(["--config-file", exp_file])

    result = cli_runner.invoke(cli.stop_job, cli_inputs)
    assert 0 == result.exit_code

    mock_stop.assert_called_once_with(klio_config, "cancel")
    mock_get_config.assert_called_once_with(exp_file)


@pytest.mark.parametrize(
    "pytest_args",
    [
        [],
        ["test_file.py::test_foo"],
        ["-s -h"],
        ["test_file.py::test_foo test:file_.py::test_foo2 -s"],
    ],
)
def test_test_job(
    pytest_args,
    mocker,
    monkeypatch,
    cli_runner,
    patch_get_config,
    patch_klio_config,
):
    mock_test = mocker.Mock()
    monkeypatch.setattr(pytest, "main", mock_test)
    mock_test.return_value = 0

    result = cli_runner.invoke(cli.test_job, pytest_args)

    core_testing.assert_execution_success(result)
    assert "true" == os.environ["KLIO_TEST_MODE"]

    mock_test.assert_called_once_with(pytest_args)


@pytest.mark.parametrize(
    "config_file_override,pytest_args",
    [(None, []), ("klio-job2.yaml", ["test_file.py::test_foo"])],
)
def test_test_job_conf_override(
    config_file_override,
    pytest_args,
    mocker,
    monkeypatch,
    config,
    cli_runner,
    patch_klio_config,
    tmpdir,
):
    mock_get_config = mocker.Mock()
    monkeypatch.setattr(cli, "_get_config", mock_get_config)
    mock_get_config.return_value = config

    mock_test = mocker.Mock()
    monkeypatch.setattr(pytest, "main", mock_test)
    mock_test.return_value = 0

    cli_inputs = []

    temp_dir = tmpdir.mkdir("testing")
    temp_dir_str = str(temp_dir)
    monkeypatch.setattr(os, "getcwd", lambda: temp_dir_str)

    config_file = "klio-job.yaml"
    if config_file_override:
        # create a tmp file else click will complain it doesn't exist
        config_file = _create_tmp_config_file(
            temp_dir_str, config_file_override, config
        )
        cli_inputs.extend(["--config-file", config_file])

    cli_inputs.extend(pytest_args)

    result = cli_runner.invoke(cli.test_job, cli_inputs)

    core_testing.assert_execution_success(result)
    assert "true" == os.environ["KLIO_TEST_MODE"]

    mock_test.assert_called_once_with(pytest_args)
    mock_get_config.assert_called_once_with(config_file)


@pytest.mark.parametrize(
    "pytest_args",
    [
        [],
        ["test_file.py::test_foo"],
        ["-s -h"],
        ["test_file.py::test_foo test:file_.py::test_foo2 -s"],
    ],
)
def test_test_job_raises(
    pytest_args,
    mocker,
    monkeypatch,
    cli_runner,
    patch_get_config,
    patch_klio_config,
):
    mock_test = mocker.Mock()
    mock_test.return_value = 1
    monkeypatch.setattr(pytest, "main", mock_test)

    result = cli_runner.invoke(cli.test_job, pytest_args)

    assert 1 == result.exit_code

    assert "true" == os.environ["KLIO_TEST_MODE"]
    mock_test.assert_called_once_with(pytest_args)


@pytest.mark.parametrize(
    "input_file,entity_ids,exp_raise,exp_msg",
    (
        ("input.txt", (), False, None),
        (None, ("foo", "bar"), False, None),
        ("input.txt", ("foo", "bar"), True, "Illegal usage"),
        (None, (), True, "Must provide"),
    ),
)
def test_require_profile_input_data(
    input_file, entity_ids, exp_raise, exp_msg
):
    if exp_raise:
        with pytest.raises(click.UsageError, match=exp_msg):
            cli._require_profile_input_data(input_file, entity_ids)
    else:
        ret = cli._require_profile_input_data(input_file, entity_ids)
        assert ret is None


@pytest.mark.parametrize("input_file", (None, "input-data.txt"))
@pytest.mark.parametrize("show_logs", (True, False))
def test_run_profile_pipeline(
    input_file,
    show_logs,
    cli_runner,
    mocker,
    patch_klio_config,
    patch_get_config,
    klio_config,
):
    mock_req_prof_input_data = mocker.patch.object(
        cli, "_require_profile_input_data"
    )
    mock_klio_pipeline = mocker.patch.object(profile, "KlioPipeline")
    mock_logging_disable = mocker.patch.object(cli.logging, "disable")

    cli_inputs = ["run-pipeline"]
    if show_logs:
        cli_inputs.append("--show-logs")

    with cli_runner.isolated_filesystem():
        if input_file:
            with open(input_file, "w") as in_f:
                in_f.write("foo\n")
            cli_inputs.extend(["--input-file", input_file])

        result = cli_runner.invoke(cli.profile_job, cli_inputs)

    assert 0 == result.exit_code

    mock_req_prof_input_data.assert_called_once_with(input_file, ())
    mock_klio_pipeline.assert_called_once_with(
        input_file=input_file, entity_ids=(), klio_config=klio_config
    )
    mock_klio_pipeline.return_value.profile.assert_called_once_with(what="run")

    if not show_logs:
        mock_logging_disable.assert_called_once_with(cli.logging.CRITICAL)
    else:
        mock_logging_disable.assert_not_called()


@pytest.mark.parametrize(
    "interval,inc_child,multiproc", ((None, False, False), (0.5, True, True))
)
@pytest.mark.parametrize(
    "plot_graph,show_logs", ((True, True), (False, False))
)
@pytest.mark.parametrize(
    "input_file,output_file", ((None, None), ("input.txt", "output.txt"))
)
@pytest.mark.parametrize("entity_ids", (None, ("foo", "bar")))
def test_profile_memory(
    entity_ids,
    input_file,
    output_file,
    plot_graph,
    show_logs,
    interval,
    inc_child,
    multiproc,
    cli_runner,
    mocker,
    patch_klio_config,
    patch_get_config,
    klio_config,
):
    mock_req_prof_input_data = mocker.patch.object(
        cli, "_require_profile_input_data"
    )
    mock_klio_pipeline = mocker.patch.object(profile, "KlioPipeline")
    if not plot_graph:
        # it otherwise returns a mock, which evaluates to True
        mock_klio_pipeline.return_value.profile.return_value = None

    cli_inputs = ["memory"]
    if interval:
        cli_inputs.extend(["--interval", interval])
    if inc_child:
        cli_inputs.append("--include-children")
    if multiproc:
        cli_inputs.append("--multiprocess")
    if plot_graph:
        cli_inputs.append("--plot-graph")
    if show_logs:
        cli_inputs.append("--show-logs")
    if output_file:
        cli_inputs.extend(["--output-file", output_file])
    if entity_ids:
        cli_inputs.extend(entity_ids)

    with cli_runner.isolated_filesystem():
        if input_file:
            with open(input_file, "w") as in_f:
                in_f.write("foo\n")
            cli_inputs.extend(["--input-file", input_file])

        result = cli_runner.invoke(cli.profile_job, cli_inputs)

    assert 0 == result.exit_code

    mock_req_prof_input_data.assert_called_once_with(
        input_file, entity_ids or ()
    )
    mock_klio_pipeline.assert_called_once_with(
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids or (),
        klio_config=klio_config,
    )
    exp_kwargs = {
        "include_children": inc_child,
        "multiprocess": multiproc,
        "interval": interval or 0.1,
        "show_logs": show_logs,
        "plot_graph": plot_graph,
    }

    mock_klio_pipeline.return_value.profile.assert_called_once_with(
        what="memory", **exp_kwargs
    )
    if plot_graph:
        assert "Memory plot graph generated at" in result.output
    else:
        assert "" == result.output


@pytest.mark.parametrize("maximum,per_element", ((True, False), (False, True)))
@pytest.mark.parametrize("show_logs", (True, False))
@pytest.mark.parametrize(
    "input_file,output_file", ((None, None), ("input.txt", "output.txt"))
)
@pytest.mark.parametrize("entity_ids", (None, ("foo", "bar")))
def test_profile_memory_per_line(
    entity_ids,
    input_file,
    output_file,
    maximum,
    per_element,
    show_logs,
    cli_runner,
    mocker,
    patch_klio_config,
    patch_get_config,
    klio_config,
):
    mock_req_prof_input_data = mocker.patch.object(
        cli, "_require_profile_input_data"
    )
    mock_klio_pipeline = mocker.patch.object(profile, "KlioPipeline")
    mock_logging_disable = mocker.patch.object(cli.logging, "disable")

    cli_inputs = ["memory-per-line"]

    if maximum:
        cli_inputs.append("--maximum")
    if per_element:
        cli_inputs.append("--per-element")
    if show_logs:
        cli_inputs.append("--show-logs")
    if output_file:
        cli_inputs.extend(["--output-file", output_file])
    if entity_ids:
        cli_inputs.extend(entity_ids)

    with cli_runner.isolated_filesystem():
        if input_file:
            with open(input_file, "w") as in_f:
                in_f.write("foo\n")
            cli_inputs.extend(["--input-file", input_file])

        result = cli_runner.invoke(cli.profile_job, cli_inputs)

    assert 0 == result.exit_code

    if not show_logs:
        mock_logging_disable.assert_called_once_with(cli.logging.CRITICAL)
    else:
        mock_logging_disable.assert_not_called()

    mock_req_prof_input_data.assert_called_once_with(
        input_file, entity_ids or ()
    )
    mock_klio_pipeline.assert_called_once_with(
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids or (),
        klio_config=klio_config,
    )

    mock_klio_pipeline.return_value.profile.assert_called_once_with(
        what="memory_per_line", get_maximum=maximum
    )


@pytest.mark.parametrize(
    "interval,plot_graph,show_logs", ((None, True, True), (0.5, False, False))
)
@pytest.mark.parametrize(
    "input_file,output_file", ((None, None), ("input.txt", "output.txt"))
)
@pytest.mark.parametrize("entity_ids", (None, ("foo", "bar")))
def test_profile_cpu(
    entity_ids,
    input_file,
    output_file,
    plot_graph,
    show_logs,
    interval,
    cli_runner,
    mocker,
    patch_klio_config,
    patch_get_config,
    klio_config,
):
    mock_req_prof_input_data = mocker.patch.object(
        cli, "_require_profile_input_data"
    )
    mock_klio_pipeline = mocker.patch.object(profile, "KlioPipeline")
    if not plot_graph:
        # it otherwise returns a mock, which evaluates to True
        mock_klio_pipeline.return_value.profile.return_value = None

    cli_inputs = ["cpu"]
    if interval:
        cli_inputs.extend(["--interval", interval])
    if plot_graph:
        cli_inputs.append("--plot-graph")
    if show_logs:
        cli_inputs.append("--show-logs")
    if output_file:
        cli_inputs.extend(["--output-file", output_file])
    if entity_ids:
        cli_inputs.extend(entity_ids)

    with cli_runner.isolated_filesystem():
        if input_file:
            with open(input_file, "w") as in_f:
                in_f.write("foo\n")
            cli_inputs.extend(["--input-file", input_file])

        result = cli_runner.invoke(cli.profile_job, cli_inputs)

    assert 0 == result.exit_code

    mock_req_prof_input_data.assert_called_once_with(
        input_file, entity_ids or ()
    )
    mock_klio_pipeline.assert_called_once_with(
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids or (),
        klio_config=klio_config,
    )
    exp_kwargs = {
        "interval": interval or 0.1,
        "show_logs": show_logs,
        "plot_graph": plot_graph,
    }

    mock_klio_pipeline.return_value.profile.assert_called_once_with(
        what="cpu", **exp_kwargs
    )
    if plot_graph:
        assert "CPU plot graph generated at" in result.output
    else:
        assert "" == result.output


@pytest.mark.parametrize("iterations,show_logs", ((None, True), (5, False)))
@pytest.mark.parametrize(
    "input_file,output_file", ((None, None), ("input.txt", "output.txt"))
)
@pytest.mark.parametrize("entity_ids", (None, ("foo", "bar")))
def test_profile_wall_time(
    entity_ids,
    input_file,
    output_file,
    iterations,
    show_logs,
    cli_runner,
    mocker,
    patch_klio_config,
    patch_get_config,
    klio_config,
):
    mock_req_prof_input_data = mocker.patch.object(
        cli, "_require_profile_input_data"
    )
    mock_klio_pipeline = mocker.patch.object(profile, "KlioPipeline")
    mock_logging_disable = mocker.patch.object(cli.logging, "disable")

    cli_inputs = ["timeit"]

    if iterations:
        cli_inputs.extend(["--iterations", str(iterations)])
    if show_logs:
        cli_inputs.append("--show-logs")
    if output_file:
        cli_inputs.extend(["--output-file", output_file])
    if entity_ids:
        cli_inputs.extend(entity_ids)

    with cli_runner.isolated_filesystem():
        if input_file:
            with open(input_file, "w") as in_f:
                in_f.write("foo\n")
            cli_inputs.extend(["--input-file", input_file])

        result = cli_runner.invoke(cli.profile_job, cli_inputs)

    assert 0 == result.exit_code

    if not show_logs:
        mock_logging_disable.assert_called_once_with(cli.logging.CRITICAL)
    else:
        mock_logging_disable.assert_not_called()

    mock_req_prof_input_data.assert_called_once_with(
        input_file, entity_ids or ()
    )
    mock_klio_pipeline.assert_called_once_with(
        input_file=input_file,
        output_file=output_file,
        entity_ids=entity_ids or (),
        klio_config=klio_config,
    )

    exp_iterations = iterations or 10
    mock_klio_pipeline.return_value.profile.assert_called_once_with(
        what="timeit", iterations=exp_iterations
    )


@pytest.mark.parametrize("config_file", (None, "klio-job2.yaml"))
@pytest.mark.parametrize("list_steps", (True, False))
def test_audit_job(
    list_steps,
    config_file,
    patch_get_config,
    cli_runner,
    tmpdir,
    mocker,
    monkeypatch,
    mock_terminal_writer,
):
    test_dir = tmpdir.mkdir("testing")
    monkeypatch.setattr(os, "getcwd", lambda: str(test_dir))
    mock_audit = mocker.Mock()
    monkeypatch.setattr(cli.audit, "audit", mock_audit)
    mock_klio_config = mocker.Mock()
    monkeypatch.setattr(cli.config, "KlioConfig", mock_klio_config)

    # mocking out what audit.list_audit_steps calls, rather than the func
    # itself since it is a little difficult to patch directly (it's a
    # callback in a decorator, where it gets evaluated at import time, not
    # runtime)
    mock_print_plugins = mocker.Mock()
    monkeypatch.setattr(
        cli.audit.plugin_utils, "print_plugins", mock_print_plugins
    )

    cli_inputs = []
    if list_steps:
        cli_inputs.append("--list")
    if config_file:
        path = test_dir.join(config_file)
        path.write("foo")
        cli_inputs.extend(["--config-file", str(path)])

    ret = cli_runner.invoke(cli.audit_job, cli_inputs)

    assert 0 == ret.exit_code
    if not list_steps:
        mock_terminal_writer.sep.assert_not_called()
        mock_print_plugins.assert_not_called()
        assert "true" == os.environ["KLIO_TEST_MODE"]
        mock_audit.assert_called_once_with(
            str(test_dir), mock_klio_config.return_value
        )
    else:
        # call args are already tested in test_audit.py::test_list_audit_steps
        assert 1 == mock_terminal_writer.sep.call_count
        # difficult to actually patch the click context that's passed into the
        # callback; let's just care that it's actually called for now
        assert 1 == mock_print_plugins.call_count
        mock_audit.assert_not_called()



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

import py
import pytest


@pytest.fixture
def caplog(caplog):
    """Set global test logging levels."""
    caplog.set_level(logging.DEBUG)
    return caplog


@pytest.fixture
def mock_terminal_writer(mocker, monkeypatch):
    mock_tw = mocker.Mock()
    monkeypatch.setattr(py.io, "TerminalWriter", mock_tw)
    return mock_tw.return_value



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import tempfile
import traceback

from klio_exec.commands.audit_steps import base


class TempFileUsage(base.BaseKlioAuditStep):
    """Avoid leaky file descriptors from `tempfile.TemporaryFile`."""

    AUDIT_STEP_NAME = "tempfile"
    PACKAGES_TO_IGNORE = ("_pytest",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tempfile_used = False
        self._tempfile_tracebacks = []

    def _mock_tempfile(self):
        """Override tempfile.TemporaryFile in the user's code
            with a MockTemporaryFile that marks the class attribute
            `TempfileUsage.AuditStep.mock_temporary_file_was used`
             as True before returning an actual tempfile.TemporaryFile.

            Ignores any use of tempfile.TemporaryFile by
            pytest.
        """
        RealTempFile = tempfile.TemporaryFile

        def MockTemporaryFile(*args, **kwargs):
            stack = traceback.extract_stack()[:-1]

            caller_frame = stack[-1]
            should_ignore = any(
                [
                    ("/%s/" % ignored) in caller_frame.filename
                    for ignored in TempFileUsage.PACKAGES_TO_IGNORE
                ]
            )
            if not should_ignore:
                self._tempfile_used = True
                self._tempfile_tracebacks.append(stack)
            return RealTempFile(*args, **kwargs)

        tempfile.TemporaryFile = MockTemporaryFile

    def before_tests(self):
        self._mock_tempfile()

    def after_tests(self):
        if self._tempfile_used:
            self.emit_error(
                "`tempfile.TemporaryFile` was used! Please use `tempfile."
                "NamedTemporaryFile` instead to avoid leaking files. "
                "Traceback:",
                self._tempfile_tracebacks[0],
            )


# shortcut for registering plugins in setup.py
_init = TempFileUsage



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.utils import wrappers


class DummyTransformGenerator(object):
    def process(self, x):
        yield x + 10
        yield x + 20


class DummyTransformGeneratorRaises(object):
    def process(self, x):
        yield x + 10
        yield x + 20
        raise Exception("catch me")


class DummyTransformFunc(object):
    def process(self, x):
        return x + 10


class DummyTransformFuncRaises(object):
    def process(self, *args):
        raise Exception("catch me")


class TestProfiler(wrappers.KLineProfilerMixin):
    def add_function(self, *args, **kwargs):
        pass

    def enable_by_count(self, *args, **kwargs):
        pass

    def disable_by_count(self, *args, **kwargs):
        pass


@pytest.fixture
def k_profiler():
    return TestProfiler()


@pytest.fixture
def mock_add_func(k_profiler, mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(k_profiler, "add_function", mock)
    return mock


@pytest.fixture
def mock_enable_by_count(k_profiler, mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(k_profiler, "enable_by_count", mock)
    return mock


@pytest.fixture
def mock_disable_by_count(k_profiler, mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(k_profiler, "disable_by_count", mock)
    return mock


def dummy_func(x):
    return x + 10


def dummy_gen(x):
    yield x + 10
    yield x + 20


@pytest.mark.parametrize(
    "transform,print_msg",
    (
        (DummyTransformGenerator, ""),
        (DummyTransformGeneratorRaises, "WARN: Error caught while profiling"),
    ),
)
def test_print_user_exceptions_generators(transform, print_msg, capsys):
    transform, *_ = list(wrappers.print_user_exceptions([transform]))

    # fails without functools.wraps(func)
    assert "process" == transform.process.__name__

    # first argument is "self" of the process method
    result = transform.process(None, 10)

    assert 20 == next(result)
    assert 30 == next(result)
    with pytest.raises(StopIteration):
        next(result)

    captured = capsys.readouterr()
    assert print_msg in captured.out


@pytest.mark.parametrize(
    "transform,print_msg,exp_ret",
    (
        (DummyTransformFunc, "", 20),
        (DummyTransformFuncRaises, "WARN: Error caught while profiling", None),
    ),
)
def test_print_user_exceptions_funcs(transform, print_msg, exp_ret, capsys):
    transform, *_ = list(wrappers.print_user_exceptions([transform]))

    # fails without functools.wraps(func)
    assert "process" == transform.process.__name__

    # first argument is "self" of the process method
    result = transform.process(None, 10)

    assert exp_ret == result
    captured = capsys.readouterr()
    assert print_msg in captured.out


def test_call(
    k_profiler, mock_add_func, mock_enable_by_count, mock_disable_by_count
):
    wrapper = k_profiler.__call__(dummy_func)

    mock_add_func.assert_called_once_with(dummy_func)
    mock_enable_by_count.assert_not_called()
    mock_disable_by_count.assert_not_called()
    assert "dummy_func" == wrapper.__name__

    result = wrapper(10)
    assert 20 == result
    mock_enable_by_count.assert_called_once_with()
    mock_disable_by_count.assert_called_once_with()

    mock_add_func.reset_mock()
    mock_enable_by_count.reset_mock()
    mock_disable_by_count.reset_mock()

    wrapper = k_profiler.__call__(dummy_gen)

    mock_add_func.assert_called_once_with(dummy_gen)
    mock_enable_by_count.assert_not_called()
    mock_disable_by_count.assert_not_called()
    assert "dummy_gen" == wrapper.__name__

    result = wrapper(10)

    assert 20 == next(result)
    assert 30 == next(result)
    with pytest.raises(StopIteration):
        next(result)

    mock_enable_by_count.assert_called_once_with()
    mock_disable_by_count.assert_called_once_with()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands.utils import memory_utils


@pytest.fixture
def k_profiler():
    return memory_utils.KMemoryLineProfiler()


@pytest.fixture
def mock_show_results(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(memory_utils.mp, "show_results", mock)
    return mock


def dummy_func(x):
    return x + 10


def dummy_gen(x):
    yield x + 10
    yield x + 20


def test_wrap_per_element(k_profiler, mock_show_results, mocker, monkeypatch):
    # don't want to mock the whole class since we need to test a class method,
    # just mocking out the creation of a new instance to return what we can
    # control
    mock_new = mocker.Mock(return_value=k_profiler)
    monkeypatch.setattr(memory_utils.KMemoryLineProfiler, "__new__", mock_new)

    wrapper = memory_utils.KMemoryLineProfiler.wrap_per_element(dummy_func)

    mock_new.assert_not_called()
    mock_show_results.assert_not_called()
    assert "dummy_func" == wrapper.__name__

    result = wrapper(10)
    assert 20 == result
    mock_new.assert_called_once_with(
        memory_utils.KMemoryLineProfiler, backend="psutil"
    )
    mock_show_results.assert_called_once_with(k_profiler, stream=None)

    mock_new.reset_mock()
    mock_show_results.reset_mock()

    wrapper = memory_utils.KMemoryLineProfiler.wrap_per_element(dummy_gen)

    mock_new.assert_not_called()
    mock_show_results.assert_not_called()
    assert "dummy_gen" == wrapper.__name__

    result = wrapper(10)
    assert 20 == next(result)
    assert 30 == next(result)
    with pytest.raises(StopIteration):
        next(result)
    mock_new.assert_called_once_with(
        memory_utils.KMemoryLineProfiler, backend="psutil"
    )
    mock_show_results.assert_called_once_with(k_profiler, stream=None)


def test_wrap_maximum(k_profiler, mocker, monkeypatch):
    wrapper = memory_utils.KMemoryLineProfiler.wrap_maximum(
        k_profiler, dummy_func
    )

    assert "dummy_func" == wrapper.__name__

    result = wrapper(10)
    assert 20 == result

    wrapper = memory_utils.KMemoryLineProfiler.wrap_maximum(
        k_profiler, dummy_gen
    )

    assert "dummy_gen" == wrapper.__name__

    result = wrapper(10)
    assert 20 == next(result)
    assert 30 == next(result)
    with pytest.raises(StopIteration):
        next(result)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click

#####
# options for `klioexec run`
#####


def config_file(func):
    return click.option(
        "-c",
        "--config-file",
        type=click.Path(exists=True, resolve_path=True, dir_okay=False),
        help=(
            "Path to config filename. Defaults to `klio-job.yaml` in the "
            "current working directory. "
        ),
    )(func)


def blocking(func):
    return click.option(
        "--blocking/--no-blocking",
        default=None,
        is_flag=True,
        help="Wait for Dataflow job to finish before returning",
    )(func)


#####
# options for `klioexec profile`
#####
def input_file(func):
    # mutually exclusive with entity ID click args
    return click.option(
        "-i",
        "--input-file",
        type=click.Path(exists=True, dir_okay=False, readable=True),
        help=(
            "File of entity IDs (separated by a new line character) with "
            "which to profile a Klio job."
        ),
        required=False,
    )(func)


def output_file(func):
    return click.option(
        "-o",
        "--output-file",
        type=click.Path(exists=False, dir_okay=False, writable=True),
        default=None,
        show_default="stdout",
        help="Output file for results.",
    )(func)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import pytest

from klio_exec.commands.utils import profile_utils


def test_get_profiling_data(mocker):
    profiling_data = "MEM 100 100001\n" "MEM 200 100002\n" "MEM 300 100003\n"
    m_open = mocker.mock_open(read_data=profiling_data)
    mock_open = mocker.patch(
        "klio_exec.commands.utils.profile_utils.open", m_open
    )

    act_ret = profile_utils._get_profiling_data("foo.txt")

    mock_open.assert_called_once_with("foo.txt", "r")
    exp_ret = {
        "data": [100.0, 200.0, 300.0],
        "timestamp": [100001.0, 100002.0, 100003.0],
    }
    assert exp_ret == act_ret


def test_get_profiling_data_raises_no_file(mocker, caplog):
    with pytest.raises(SystemExit):
        ret = profile_utils._get_profiling_data("i_should_not_exist.txt")
        assert ret is None

    assert 1 == len(caplog.records)
    assert "ERROR" == caplog.records[0].levelname
    assert "Could not read profiling data" in caplog.records[0].message


def test_get_profiling_data_malformed_data(mocker):
    profiling_data = (
        "MEM 100 100001\n" "foobar baz\n" "\n" "MEM 1 2 3\n" "MEM 300 100003\n"
    )
    m_open = mocker.mock_open(read_data=profiling_data)
    mock_open = mocker.patch(
        "klio_exec.commands.utils.profile_utils.open", m_open
    )

    act_ret = profile_utils._get_profiling_data("foo.txt")

    mock_open.assert_called_once_with("foo.txt", "r")
    exp_ret = {"data": [100.0, 300.0], "timestamp": [100001.0, 100003.0]}
    assert exp_ret == act_ret


def test_get_profiling_data_raises_no_data(mocker, caplog):
    m_open = mocker.mock_open(read_data="\n")
    mock_open = mocker.patch(
        "klio_exec.commands.utils.profile_utils.open", m_open
    )

    with pytest.raises(SystemExit):
        ret = profile_utils._get_profiling_data("foo.txt")
        assert ret is None

    mock_open.assert_called_once_with("foo.txt", "r")
    assert 1 == len(caplog.records)
    assert "ERROR" == caplog.records[0].levelname
    assert "No samples to parse in" in caplog.records[0].message


def test_plot(mocker, monkeypatch):
    parsed_data = {
        "data": [100.0, 200.0, 300.0],
        "timestamp": [100001.0, 100002.0, 100003.0],
    }
    mock_get_profiling_data = mocker.Mock()
    mock_get_profiling_data.return_value = parsed_data
    monkeypatch.setattr(
        profile_utils, "_get_profiling_data", mock_get_profiling_data
    )
    mock_plt = mocker.Mock()
    monkeypatch.setattr(profile_utils, "plt", mock_plt)

    profile_utils.plot(
        "input.txt", "output.png", "x-label", "y-label", "title test"
    )

    mock_get_profiling_data.assert_called_once_with("input.txt")
    mock_plt.figure.assert_called_once_with(figsize=(14, 6), dpi=90)

    # can't directly compare (or assert called with) np arrays;
    # grabbing the args/kwargs then using np.allclose with a 0 tolerance
    exp_data = np.asarray(parsed_data["data"])
    ts = np.asarray(parsed_data["timestamp"])
    exp_t = ts - float(ts[0])
    act_call_args, act_call_kwargs = mock_plt.plot.call_args
    assert 3 == len(act_call_args)
    assert not act_call_kwargs
    assert np.allclose(exp_t, act_call_args[0], rtol=0)
    assert np.allclose(exp_data, act_call_args[1], rtol=0)
    assert "+-c" == act_call_args[2]

    mock_plt.xlabel.assert_called_once_with("x-label")
    mock_plt.ylabel.assert_called_once_with("y-label")
    mock_plt.title.assert_called_once_with("title test")
    mock_plt.grid.assert_called_once_with()
    mock_plt.savefig.assert_called_once_with("output.png")



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

try:
    import matplotlib.pyplot as plt
except ImportError:  # pragma: no cover
    # can be a problem on MacOS w certain Python & OS versions
    # https://stackoverflow.com/a/34583958/1579977
    import matplotlib

    matplotlib.use("agg")
    import matplotlib.pyplot as plt
import numpy as np


def _get_profiling_data(filename):
    """Read a given file and parse its content for profiling data."""
    data, timestamps = [], []

    try:
        with open(filename, "r") as f:
            file_data = f.readlines()
    except Exception:
        logging.error("Could not read profiling data.", exc_info=True)
        raise SystemExit(1)

    for line in file_data:
        if line == "\n":
            continue

        line = line.strip()
        line_data = line.split(" ")
        if len(line_data) != 3:
            continue
        _, mem_usage, timestamp = line.split(" ")
        data.append(float(mem_usage))
        timestamps.append(float(timestamp))

    if not data:
        logging.error("No samples to parse in {}.".format(filename))
        raise SystemExit(1)

    return {"data": data, "timestamp": timestamps}


def plot(input_file, output_file, x_label, y_label, title):
    """Plot profiling data."""
    profile_data = _get_profiling_data(input_file)

    data = np.asarray(profile_data["data"])
    timestamp = np.asarray(profile_data["timestamp"])

    global_start = float(timestamp[0])
    t = timestamp - global_start  # start at 0 rather than a specific time

    plt.figure(figsize=(14, 6), dpi=90)
    plt.plot(t, data, "+-c")  # c is for `cyan`
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    plt.grid()
    plt.savefig(output_file)



# Copyright 2019-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import datetime
import logging
import time

from googleapiclient import discovery

#####
# TODO: this is nearly identical to klio_cli/commands/stop_job.py. This is
#       copy-pasta'ed to avoid depending on `klio-cli` and having any weird
#       dependency version conflicts for now. Ideally, the copied code should
#       be extracted out into a light-weight shared lib.  @lynn
#####


JOB_STATE_MAP = {"cancel": "JOB_STATE_CANCELLED", "drain": "JOB_STATE_DRAINED"}


_client = None


def _set_dataflow_client(api_version=None):
    global _client
    if not api_version:
        api_version = "v1b3"
    _client = discovery.build("dataflow", api_version)


def _check_job_running(config):
    request = (
        _client.projects()
        .locations()
        .jobs()
        .list(
            projectId=config.pipeline_options.project,
            location=config.pipeline_options.region,
            filter="ACTIVE",
        )
    )

    try:
        response = request.execute()
    except Exception as e:
        logging.warning(
            "Could not find running job '{}' in project '{}': {}".format(
                config.job_name, config.pipeline_options.project, e
            )
        )
        logging.warning(
            "Continuing to attempt deploying '{}'".format(config.job_name)
        )
        return

    job_results = response.get("jobs", [])
    if job_results:
        for result in job_results:
            if result["name"] == config.job_name:
                return result


def _update_job_state(job, req_state=None, retries=None):
    if retries is None:
        retries = 0

    _req_state = JOB_STATE_MAP.get(req_state, JOB_STATE_MAP["cancel"])
    if job.get("requestedState") is not _req_state:
        job["requestedState"] = _req_state

    request = (
        _client.projects()
        .locations()
        .jobs()
        .update(
            jobId=job["id"],
            projectId=job["projectId"],
            location=job["location"],
            body=job,
        )
    )

    try:
        request.execute()

    except Exception as e:
        # generic catch if 4xx error - probably shouldn't retry
        if getattr(e, "resp", None):
            if e.resp.status < 500:
                msg = "Failed to {} job '{}': {}".format(
                    req_state, job["name"], e
                )
                logging.error(msg)
                raise SystemExit(1)

        if retries > 2:
            msg = "Max retries reached: could not {} job '{}': {}".format(
                req_state, job["name"], e
            )
            logging.error(msg)
            raise SystemExit(1)

        logging.info(
            "Failed to {} job '{}'. Trying again after 30s...".format(
                req_state, job["name"]
            )
        )
        retries += 1
        time.sleep(30)
        _update_job_state(job, req_state, retries)


def _watch_job_state(job, timeout=600):
    timeout_end = datetime.datetime.now() + datetime.timedelta(seconds=timeout)

    request = (
        _client.projects()
        .locations()
        .jobs()
        .get(
            jobId=job["id"],
            projectId=job["projectId"],
            location=job["location"],
        )
    )

    while datetime.datetime.now() < timeout_end:
        try:
            resp = request.execute()
        except Exception as e:
            msg = (
                "Failed to get current status for job '{}'. Error: {}.\n"
                "Trying again after 5s...".format(job["name"], e)
            )
            logging.info(msg)
            time.sleep(5)
            continue

        if resp["currentState"] in JOB_STATE_MAP.values():
            return
        else:
            msg = "Waiting for job '{}' to reach a terminal state...".format(
                job["name"]
            )
            logging.info(msg)
            time.sleep(5)

    msg = "Job '{}' did not reach a terminal state after '{}' seconds.".format(
        job["name"], timeout
    )
    logging.error(msg)
    raise SystemExit(1)


def stop(config, strategy):
    _set_dataflow_client()
    current_running_job = _check_job_running(config)

    if not current_running_job:
        logging.info("Found no currently running job to stop.")
        return

    _update_job_state(current_running_job, req_state=strategy)
    _watch_job_state(current_running_job)
    verb = "cancelled" if strategy == "cancel" else "drained"
    logging.info("Successfully {} job '{}'".format(verb, config.job_name))



# Copyright 2021 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import warnings

from apache_beam import pipeline as beam_pipeline
from apache_beam.options import pipeline_options
from apache_beam.options import value_provider
from apache_beam.runners.direct import bundle_factory
from apache_beam.runners.direct import clock as beam_clock
from apache_beam.runners.direct import (
    consumer_tracking_pipeline_visitor as ctpv,
)
from apache_beam.runners.direct import direct_runner
from apache_beam.runners.direct import evaluation_context as eval_ctx
from apache_beam.runners.direct import executor as beam_exec
from apache_beam.testing import test_stream

from klio_exec.runners import evaluators


# without this, users would get flooded with warnings of "your application has
# authenticated using end user credentials from Google Cloud SDK ..."
warnings.filterwarnings(
    "ignore", category=UserWarning, module="google.auth._default",
)
_LOGGER = logging.getLogger("klio.gke_direct_runner")


class GkeDirectRunner(direct_runner.BundleBasedDirectRunner):
    """Custom DirectRunner class for running on GKE.

    Acknowledges PubsubMessages after they are finished processing rather than
    before they start processing, but otherwise is meant to behave identically
    to the BundleBasedDirectRunner.
    """

    def run_pipeline(self, pipeline, options):
        """Execute the entire pipeline and returns an DirectPipelineResult."""

        # Klio maintainer note: This code is the eact same logic in
        # direct_runner.BundleBasedDirectRunner.run_pipeline with the
        # following changes:
        # 1. Import statements that were originally inside this method
        #    was moved to the top of this module.
        # 2. Import statements adjusted to import module and not objects
        #    according to the google style guide.
        # 3. The functionalty we needed to override, which is invoking
        #    our own TransformEvaluatorRegistry when instantiating the
        #    Executor class (called out below).

        # If the TestStream I/O is used, use a mock test clock.
        class TestStreamUsageVisitor(beam_pipeline.PipelineVisitor):
            """Visitor determining whether a Pipeline uses a TestStream."""

            def __init__(self):
                self.uses_test_stream = False

            def visit_transform(self, applied_ptransform):
                if isinstance(
                    applied_ptransform.transform, test_stream.TestStream
                ):
                    self.uses_test_stream = True

        visitor = TestStreamUsageVisitor()
        pipeline.visit(visitor)
        clock = (
            beam_clock.TestClock()
            if visitor.uses_test_stream
            else beam_clock.RealClock()
        )

        # Performing configured PTransform overrides.
        pipeline.replace_all(direct_runner._get_transform_overrides(options))

        _LOGGER.info("Running pipeline with Klio's GkeDirectRunner.")
        self.consumer_tracking_visitor = ctpv.ConsumerTrackingPipelineVisitor()
        pipeline.visit(self.consumer_tracking_visitor)

        bndl_factory = bundle_factory.BundleFactory(
            stacked=options.view_as(
                pipeline_options.DirectOptions
            ).direct_runner_use_stacked_bundle
        )
        evaluation_context = eval_ctx.EvaluationContext(
            options,
            bndl_factory,
            self.consumer_tracking_visitor.root_transforms,
            self.consumer_tracking_visitor.value_to_consumers,
            self.consumer_tracking_visitor.step_names,
            self.consumer_tracking_visitor.views,
            clock,
        )

        # Klio maintainer note: this is where the change in logic is:
        # using our own `KlioTransformEvaluatorRegistry`.
        executor = beam_exec.Executor(
            self.consumer_tracking_visitor.value_to_consumers,
            evaluators.KlioTransformEvaluatorRegistry(evaluation_context),
            evaluation_context,
        )
        # DirectRunner does not support injecting
        # PipelineOptions values at runtime
        value_provider.RuntimeValueProvider.set_runtime_options({})
        # Start the executor. This is a non-blocking call, it will start the
        # execution in background threads and return.
        executor.start(self.consumer_tracking_visitor.root_transforms)
        result = direct_runner.DirectPipelineResult(
            executor, evaluation_context
        )

        return result



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import platform

from packaging import version

from klio_exec.commands.audit_steps import base


class NumPyBrokenBLASUsage(base.BaseKlioAuditStep):
    """Detect numpy version for potential threading issues."""

    AUDIT_STEP_NAME = "numpy_broken_blas"
    MINIMUM_NUMPY_VERSION = version.parse("1.16.3")

    @staticmethod
    def get_description():
        # tmp turn off black formatting to skip long URL
        # fmt: off
        return (
            "The 1.16.3 version of numpy links against a newer version of "
            "OpenBLAS that fixes some important threading issues - notably, "
            "the `dot` function that calls into OpenBLAS' _dgemv function, "
            "which on older versions, is non-reentrant and can cause both "
            "incorrect results and deadlocks.\n\n"
            "See:\n"
            "\t- https://github.com/numpy/numpy/blob/2f70544179e24b0ebc0263111f36e6450bbccf94/doc/source/release/1.16.3-notes.rst#numpy-1163-release-notes\n"  # noqa: E501
            "\t- https://github.com/xianyi/OpenBLAS/issues/1844\n"
            "\t- https://github.com/numpy/numpy/issues/12394\n"
            "\t- https://github.com/xianyi/OpenBLAS/pull/1865\n"
        )
        # fmt: on

    @property
    def _is_job_single_threaded_per_container(self):
        exps = self.klio_config.pipeline_options.experiments
        return "worker_threads=1" in exps

    @staticmethod
    def _get_current_numpy_version():
        try:
            import numpy

            return version.parse(numpy.version.short_version)
        except ImportError:
            return None

    def after_tests(self):
        if self._is_job_single_threaded_per_container:
            return

        if platform.system().lower() != "linux":
            return

        numpy_version = self._get_current_numpy_version()
        if numpy_version is None:
            return

        if numpy_version < NumPyBrokenBLASUsage.MINIMUM_NUMPY_VERSION:
            msg = (
                "Multiple threads are used, but a NumPy version older than %s "
                "was detected. Older versions of NumPy are known to be "
                "thread-unsafe due to a broken OpenBLAS dependency on Linux."
            ) % NumPyBrokenBLASUsage.MINIMUM_NUMPY_VERSION
            self.emit_error(msg)


_init = NumPyBrokenBLASUsage



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

import pytest

from klio_exec.commands import profile as profile_cmd


@pytest.fixture
def config(mocker):
    return mocker.Mock()


@pytest.fixture
def klio_pipeline(config):
    return profile_cmd.KlioPipeline(config)


class DummyTransformGenerator(object):
    def process(self, x):
        yield x + 10
        yield x + 20


class DummyTransformFunc(object):
    def process(self, x):
        return x + 10


class DummyTransformFuncRaises(object):
    def process(self, *args):
        raise Exception("catch me")


@pytest.fixture
def transforms():
    return [
        DummyTransformGenerator,
        DummyTransformFunc,
        DummyTransformFuncRaises,
    ]


@pytest.mark.parametrize("filename", (None, "-", "input.txt"))
def test_smart_open(filename, mocker):
    m_open = mocker.mock_open()
    mock_open = mocker.patch("klio_exec.commands.profile.open", m_open)

    with profile_cmd.smart_open(filename, fmode="r") as act_ret:
        pass

    if filename and filename != "-":
        mock_open.assert_called_once_with(filename, "r")
        assert act_ret.closed
    else:
        mock_open.assert_not_called()
        assert act_ret == sys.stdout


@pytest.mark.parametrize(
    "what,output_file,temp_out,exp_ret",
    (
        ("mem", "output.txt", False, "output.png"),
        ("mem", "output.txt", True, "klio_profile_mem_1234.png"),
        ("mem", "output", False, "output.png"),
    ),
)
def test_get_output_png_file(
    what, output_file, temp_out, exp_ret, klio_pipeline, monkeypatch
):
    monkeypatch.setattr(klio_pipeline, "_now_str", "1234")
    monkeypatch.setattr(klio_pipeline, "output_file", output_file)

    act_ret = klio_pipeline._get_output_png_file(what, temp_out)
    assert exp_ret == act_ret


@pytest.mark.parametrize(
    "output_file,plot_graph,exp_temp_output",
    ((None, True, True), (None, False, False), ("output.txt", True, False)),
)
def test_smart_temp_create(
    output_file,
    plot_graph,
    exp_temp_output,
    klio_pipeline,
    mocker,
    monkeypatch,
):
    monkeypatch.setattr(klio_pipeline, "output_file", output_file)
    mock_named_temp_file = mocker.Mock()
    mock_file = mocker.Mock()
    mock_file.name = "temp-1234"
    mock_named_temp_file.return_value = mock_file
    monkeypatch.setattr(
        profile_cmd.tempfile, "NamedTemporaryFile", mock_named_temp_file
    )

    with klio_pipeline._smart_temp_create(
        "mem", plot_graph
    ) as act_temp_output:
        pass

    assert exp_temp_output is act_temp_output
    if exp_temp_output:
        assert klio_pipeline.output_file == "temp-1234"


@pytest.mark.parametrize(
    "show_logs,input_file", ((False, None), (True, "input.txt"))
)
def test_get_subproc(
    show_logs, input_file, klio_pipeline, mocker, monkeypatch
):
    mock_subproc_open = mocker.Mock()
    monkeypatch.setattr(profile_cmd.subprocess, "Popen", mock_subproc_open)
    monkeypatch.setattr(klio_pipeline, "input_file", input_file)
    monkeypatch.setattr(klio_pipeline, "entity_ids", ("foo", "bar"))

    kwargs = {"show_logs": show_logs}
    klio_pipeline._get_subproc(**kwargs)

    exp_call_args = ["klioexec", "profile", "run-pipeline"]
    if show_logs:
        exp_call_args.append("--show-logs")
    if input_file:
        exp_call_args.extend(["--input-file", input_file])
    else:
        exp_call_args.extend(["foo", "bar"])

    mock_subproc_open.assert_called_once_with(exp_call_args)


@pytest.mark.parametrize("output_file", (None, "output.txt"))
def test_profile_wall_time_per_line(
    output_file, klio_pipeline, mocker, monkeypatch
):
    monkeypatch.setattr(klio_pipeline, "output_file", output_file)
    mock_line_prof = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_cpu_line_profiler", mock_line_prof
    )
    mock_run_pipeline = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_run_pipeline", mock_run_pipeline)

    klio_pipeline._profile_wall_time_per_line(iterations=1)

    if output_file:
        mock_line_prof.return_value.print_stats.assert_called_once_with(
            klio_pipeline.output_file, output_unit=1
        )
    else:
        mock_line_prof.return_value.print_stats.assert_called_once_with(
            output_unit=1
        )


@pytest.mark.parametrize("get_maximum,exp_fmode", ((True, "w"), (False, "a")))
def test_profile_memory_per_line(
    get_maximum, exp_fmode, klio_pipeline, mocker, monkeypatch
):

    mock_memory_profiler = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_memory_line_profiler", mock_memory_profiler
    )

    mock_memory_wrapper = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_memory_line_wrapper", mock_memory_wrapper
    )

    mock_smart_open = mocker.patch.object(profile_cmd, "smart_open")
    mock_smart_open.return_value.__enter__.return_value = "opened_file"

    mock_run_pipeline = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_run_pipeline", mock_run_pipeline)

    mock_show_results = mocker.Mock()
    monkeypatch.setattr(
        profile_cmd.memory_profiler, "show_results", mock_show_results
    )

    klio_pipeline._profile_memory_per_line(get_maximum=get_maximum)

    mock_memory_profiler.assert_called_once_with()
    mock_memory_wrapper.assert_called_once_with(
        mock_memory_profiler.return_value, get_maximum
    )

    mock_smart_open.assert_called_once_with(
        klio_pipeline.output_file, fmode=exp_fmode
    )
    assert "opened_file" == klio_pipeline._stream
    mock_run_pipeline.assert_called_once_with()
    if get_maximum:
        mock_show_results.assert_called_once_with(
            mock_memory_profiler.return_value, stream="opened_file"
        )
    else:
        mock_show_results.assert_not_called()


@pytest.mark.parametrize("plot_graph", (True, False))
def test_profile_memory(plot_graph, klio_pipeline, mocker, monkeypatch):
    mock_get_subproc = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_get_subproc", mock_get_subproc)
    mock_smart_temp_create = mocker.patch.object(
        klio_pipeline, "_smart_temp_create"
    )
    mock_smart_temp_create.return_value.__enter__.return_value = True
    mock_smart_open = mocker.patch.object(profile_cmd, "smart_open")
    mock_smart_open.return_value.__enter__.return_value.name = "bar"
    mock_memory_usage = mocker.Mock()
    monkeypatch.setattr(
        profile_cmd.memory_profiler, "memory_usage", mock_memory_usage
    )
    mock_get_output_png = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_output_png_file", mock_get_output_png
    )
    mock_plot = mocker.Mock()
    monkeypatch.setattr(profile_cmd.profile_utils, "plot", mock_plot)

    kwargs = {"plot_graph": plot_graph}
    act_ret = klio_pipeline._profile_memory(**kwargs)
    if plot_graph:
        mock_get_output_png.assert_called_once_with("memory", True)
        assert mock_get_output_png.return_value == act_ret
    else:
        assert act_ret is None


@pytest.mark.parametrize("plot_graph", (True, False))
def test_profile_cpu(plot_graph, klio_pipeline, mocker, monkeypatch):
    mock_get_subproc = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_get_subproc", mock_get_subproc)
    mock_smart_temp_create = mocker.patch.object(
        klio_pipeline, "_smart_temp_create"
    )
    mock_smart_temp_create.return_value.__enter__.return_value = True
    mock_smart_open = mocker.patch.object(profile_cmd, "smart_open")
    mock_smart_open.return_value.__enter__.return_value.name = "bar"
    mock_get_cpu_usage = mocker.Mock()
    monkeypatch.setattr(
        profile_cmd.cpu_utils, "get_cpu_usage", mock_get_cpu_usage
    )
    mock_get_output_png = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_output_png_file", mock_get_output_png
    )
    mock_plot = mocker.Mock()
    monkeypatch.setattr(profile_cmd.profile_utils, "plot", mock_plot)

    kwargs = {"plot_graph": plot_graph}
    act_ret = klio_pipeline._profile_cpu(**kwargs)
    if plot_graph:
        mock_get_output_png.assert_called_once_with("cpu", True)
        assert mock_get_output_png.return_value == act_ret
    else:
        assert act_ret is None


@pytest.mark.parametrize("input_file", (None, "input.txt"))
def test_get_io_mapper(
    input_file, klio_pipeline, transforms, mocker, monkeypatch
):
    entity_ids = ["id1", "id2", "id3"]
    serialized_messages = [
        klio_pipeline._entity_id_to_message(i).SerializeToString()
        for i in entity_ids
    ]

    if input_file:
        monkeypatch.setattr(klio_pipeline, "input_file", input_file)
    else:
        monkeypatch.setattr(klio_pipeline, "entity_ids", entity_ids)

    mock_read_from_text = mocker.Mock()
    monkeypatch.setattr(
        profile_cmd.beam.io, "ReadFromText", mock_read_from_text
    )
    mock_create = mocker.Mock()
    monkeypatch.setattr(profile_cmd.beam, "Create", mock_create)

    mock_flatmap = mocker.Mock()
    mock_transform = mocker.MagicMock()
    mock_flatmap.return_value = mock_transform
    monkeypatch.setattr(profile_cmd.beam, "FlatMap", mock_flatmap)

    mapper = klio_pipeline._get_io_mapper(1)

    assert isinstance(mapper, profile_cmd.StubIOMapper)

    if input_file:
        mock_read_from_text.assert_called_once_with(klio_pipeline.input_file)
    else:
        mock_create.assert_called_once_with(serialized_messages)


def test_run_pipeline(klio_pipeline, mocker, monkeypatch):

    mock_get_io_mapper = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_get_io_mapper", mock_get_io_mapper)

    mock_get_user_pipeline = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_user_pipeline", mock_get_user_pipeline
    )

    mock_get_user_config = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_get_user_config", mock_get_user_config
    )

    klio_pipeline._run_pipeline()

    mock_get_io_mapper.assert_called_once_with(1)
    mock_get_user_config.assert_called_once_with()
    mock_get_user_pipeline.assert_called_once_with(
        mock_get_user_config.return_value, mock_get_io_mapper.return_value
    )
    mock_get_user_pipeline.return_value.run.assert_called_once_with()


@pytest.mark.parametrize(
    "what", ("run", "cpu", "timeit", "memory", "memory_per_line", "foo")
)
def test_profile(what, klio_pipeline, mocker, monkeypatch):
    mock_run_pipeline = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_run_pipeline", mock_run_pipeline)

    mock_profile_cpu = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_profile_cpu", mock_profile_cpu)

    mock_profile_memory = mocker.Mock()
    monkeypatch.setattr(klio_pipeline, "_profile_memory", mock_profile_memory)

    mock_memory_per_line = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_profile_memory_per_line", mock_memory_per_line
    )

    mock_wall_time_per_line = mocker.Mock()
    monkeypatch.setattr(
        klio_pipeline, "_profile_wall_time_per_line", mock_wall_time_per_line
    )

    klio_pipeline.profile(what)

    if what == "run":
        mock_run_pipeline.assert_called_once_with()
        mock_profile_cpu.assert_not_called()
        mock_profile_memory.assert_not_called()
        mock_memory_per_line.assert_not_called()
        mock_wall_time_per_line.assert_not_called()

    elif what == "cpu":
        mock_run_pipeline.assert_not_called()
        mock_profile_cpu.assert_called_once_with()
        mock_profile_memory.assert_not_called()
        mock_memory_per_line.assert_not_called()
        mock_wall_time_per_line.assert_not_called()

    elif what == "memory":
        mock_run_pipeline.assert_not_called()
        mock_profile_cpu.assert_not_called()
        mock_profile_memory.assert_called_once_with()
        mock_memory_per_line.assert_not_called()
        mock_wall_time_per_line.assert_not_called()

    elif what == "memory_per_line":
        mock_run_pipeline.assert_not_called()
        mock_profile_cpu.assert_not_called()
        mock_profile_memory.assert_not_called()
        mock_memory_per_line.assert_called_once_with()
        mock_wall_time_per_line.assert_not_called()

    elif what == "timeit":
        mock_run_pipeline.assert_not_called()
        mock_profile_cpu.assert_not_called()
        mock_profile_memory.assert_not_called()
        mock_memory_per_line.assert_not_called()
        mock_wall_time_per_line.assert_called_once_with()

    else:
        mock_run_pipeline.assert_not_called()
        mock_profile_cpu.assert_not_called()
        mock_profile_memory.assert_not_called()
        mock_memory_per_line.assert_not_called()
        mock_wall_time_per_line.assert_not_called()



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections
import contextlib
import functools
import logging
import os
import subprocess
import sys
import tempfile
import time

import apache_beam as beam

try:
    import memory_profiler
except ImportError:  # pragma: no cover
    logging.error(
        "Failed to import profiling dependencies. Did you install "
        "`klio-exec[debug]` in your job's Docker image?"
    )
    raise SystemExit(1)

from klio.transforms import decorators
from klio_core.proto import klio_pb2

from klio_exec.commands.utils import cpu_utils
from klio_exec.commands.utils import memory_utils
from klio_exec.commands.utils import profile_utils


@contextlib.contextmanager
def smart_open(filename=None, fmode=None):
    """Handle both stdout and files in the same manner."""
    if filename and filename != "-":
        fh = open(filename, fmode)
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()


class StubIOSubMapper(object):
    def __init__(self, input_pcol):
        def fake_constructor(*args, **kwargs):
            return input_pcol

        # normally this is a map of io-name -> transform class.  Instead we'll
        # just have every possible name return our pretend constructor that
        # returns our pre-constructed transform
        self.input = collections.defaultdict(lambda: fake_constructor)
        self.output = {}  # no outputs


class StubIOMapper(object):
    def __init__(self, input_pcol, iterations):

        repeated_input = input_pcol | beam.FlatMap(lambda x: [x] * iterations)

        self.batch = StubIOSubMapper(repeated_input)
        self.streaming = StubIOSubMapper(repeated_input)

    @staticmethod
    def from_input_file(file_path, iterations):
        transform = beam.io.ReadFromText(file_path)
        return StubIOMapper(transform, iterations)

    @staticmethod
    def from_entity_ids(id_list, iterations):
        transform = beam.Create(id_list)
        return StubIOMapper(transform, iterations)


class KlioPipeline(object):
    DEFAULT_FILE_PREFIX = "klio_profile_{what}_{ts}"
    TRANSFORMS_PATH = "./transforms.py"

    def __init__(
        self, klio_config, input_file=None, output_file=None, entity_ids=None
    ):
        self.input_file = input_file
        self.output_file = output_file
        self.entity_ids = entity_ids
        self._stream = None
        self._now_str = time.strftime("%Y%m%d%H%M%S", time.localtime())
        self.klio_config = klio_config

    def _get_output_png_file(self, what, temp_output):
        output_file_base = self.output_file
        prefix = KlioPipeline.DEFAULT_FILE_PREFIX.format(
            what=what, ts=self._now_str
        )
        if temp_output:
            output_file_base = prefix

        elif "." in self.output_file:
            # reuse a user's output file name, just replace existing extension
            output_file_base = os.path.splitext(self.output_file)[0]

        return "{}.png".format(output_file_base)

    @contextlib.contextmanager
    def _smart_temp_create(self, what, plot_graph):
        # For plotting a graph, an output file of the data collected is
        # needed, but the user shouldn't be required to provide an output
        # file if they don't want. This creates a tempfile to write data
        # to for generating the plot graph off of.
        # A context manager needed so that temp file can be cleaned up after.
        temp_output = False
        prefix = KlioPipeline.DEFAULT_FILE_PREFIX.format(
            what=what, ts=self._now_str
        )
        if plot_graph and not self.output_file:
            temp_output_file = tempfile.NamedTemporaryFile(
                dir=".", prefix=prefix
            )
            self.output_file = temp_output_file.name
            temp_output = True
        yield temp_output

    def _get_subproc(self, **kwargs):
        cmd = ["klioexec", "profile", "run-pipeline"]

        if kwargs.get("show_logs"):
            cmd.append("--show-logs")

        if self.input_file:
            cmd.extend(["--input-file", self.input_file])
        else:
            cmd.extend(self.entity_ids)

        return subprocess.Popen(cmd)

    def _get_cpu_line_profiler(self):
        return cpu_utils.KLineProfiler()

    def _profile_wall_time_per_line(self, iterations, **_):
        profiler = self._get_cpu_line_profiler()
        decorators.ACTIVE_PROFILER = profiler

        self._run_pipeline(iterations=iterations)

        if self.output_file:
            return profiler.print_stats(self.output_file, output_unit=1)

        # output_unit = 1 second, meaning the numbers in "Time" and
        # "Per Hit" columns are in seconds
        profiler.print_stats(output_unit=1)

    def _get_memory_line_profiler(self):
        return memory_utils.KMemoryLineProfiler(backend="psutil")

    def _get_memory_line_wrapper(self, profiler, get_maximum):
        wrapper = memory_utils.KMemoryLineProfiler.wrap_per_element
        if get_maximum:
            wrapper = functools.partial(
                memory_utils.KMemoryLineProfiler.wrap_maximum, profiler
            )
        return wrapper

    def _profile_memory_per_line(self, get_maximum=False):
        profiler = self._get_memory_line_profiler()
        decorators.ACTIVE_PROFILER = self._get_memory_line_wrapper(
            profiler, get_maximum
        )

        # "a"ppend if output per element; "w"rite (once) for maximum.
        # append will append a file with potentially already-existing data
        # (i.e. from a previous run), which may be confusing; but with how
        # memory_profiler treats streams, there's no simple way to prevent
        # appending data for per-element without re-implementing parts of
        # memory_profiler (maybe someday?) @lynn
        fmode = "w" if get_maximum else "a"

        with smart_open(self.output_file, fmode=fmode) as f:
            self._stream = f
            self._run_pipeline()

            if get_maximum:
                memory_profiler.show_results(profiler, stream=self._stream)

    def _profile_memory(self, **kwargs):
        # Profile the memory while the pipeline runs in another process
        p = self._get_subproc(**kwargs)

        plot_graph = kwargs.get("plot_graph")
        with self._smart_temp_create("memory", plot_graph) as temp_output:
            with smart_open(self.output_file, fmode="w") as f:
                memory_profiler.memory_usage(
                    proc=p,
                    interval=kwargs.get("interval"),
                    timestamps=True,
                    include_children=kwargs.get("include_children"),
                    multiprocess=kwargs.get("multiprocess"),
                    stream=f,
                )

            if not plot_graph:
                return

            output_png = self._get_output_png_file("memory", temp_output)
            profile_utils.plot(
                input_file=self.output_file,
                output_file=output_png,
                x_label="Time (in seconds)",
                y_label="Memory used (in MiB)",
                title="Memory Used While Running Klio-based Transforms",
            )
            return output_png

    def _profile_cpu(self, **kwargs):
        # Profile the CPU while the pipeline runs in another process
        p = self._get_subproc(**kwargs)

        plot_graph = kwargs.get("plot_graph")
        with self._smart_temp_create("cpu", plot_graph) as temp_output:
            with smart_open(self.output_file, fmode="w") as f:
                cpu_utils.get_cpu_usage(
                    proc=p, interval=kwargs.get("interval"), stream=f,
                )

            if not plot_graph:
                return

            output_png = self._get_output_png_file("cpu", temp_output)
            profile_utils.plot(
                input_file=self.output_file,
                output_file=output_png,
                x_label="Time (in seconds)",
                y_label="CPU%",
                title="CPU Usage of All Klio-based Transforms",
            )
            return output_png

    def _get_user_pipeline(self, config, io_mapper):
        runtime_config = collections.namedtuple(
            "RuntimeConfig",
            ["image_tag", "direct_runner", "update", "blocking"],
        )(None, True, False, True)

        from klio_exec.commands.run import KlioPipeline as KP

        return KP("profile_job", config, runtime_config, io_mapper)

    def _get_user_config(self):
        self.klio_config.pipeline_options.runner = "direct"
        self.klio_config.job_config.events.outputs = {}
        return self.klio_config

    @staticmethod
    def _entity_id_to_message(entity_id):
        message = klio_pb2.KlioMessage()
        message.data.element = bytes(entity_id, "UTF-8")
        message.metadata.intended_recipients.anyone.SetInParent()
        message.version = klio_pb2.Version.V2
        return message

    def _get_io_mapper(self, iterations):
        if self.input_file:
            return StubIOMapper.from_input_file(self.input_file, iterations)
        else:
            messages = []
            for entity_id in self.entity_ids:
                message = self._entity_id_to_message(entity_id)
                messages.append(message.SerializeToString())
            return StubIOMapper.from_entity_ids(messages, iterations)

    def _run_pipeline(self, iterations=None, **_):
        if not iterations:
            iterations = 1

        io_mapper = self._get_io_mapper(iterations)
        config = self._get_user_config()

        pipeline = self._get_user_pipeline(config, io_mapper)
        pipeline.run()

    def profile(self, what, **kwargs):

        if what == "run":
            return self._run_pipeline(**kwargs)
        elif what == "cpu":
            return self._profile_cpu(**kwargs)
        elif what == "memory":
            return self._profile_memory(**kwargs)
        elif what == "memory_per_line":
            return self._profile_memory_per_line(**kwargs)
        elif what == "timeit":
            return self._profile_wall_time_per_line(**kwargs)



# Copyright 2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from klio_exec.commands import audit


@pytest.fixture
def mock_pytest(mocker, monkeypatch):
    mock_pt = mocker.Mock()
    monkeypatch.setattr(audit, "pytest", mock_pt)
    return mock_pt


@pytest.fixture
def mock_get_audit_steps(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr(audit, "_get_audit_steps", mock)
    return mock


@pytest.fixture
def mock_run_pytest(mocker, monkeypatch):
    mock = mocker.Mock(return_value=False)
    monkeypatch.setattr(audit, "_run_pytest", mock)
    return mock


@pytest.mark.parametrize(
    "exit_code,exp_pytest_failed,exp_msg,exp_kwargs",
    (
        (1, True, "PyTest failed!\n", {"yellow": True}),
        (0, False, "PyTest passed!\n", {}),
    ),
)
def test_run_pytest(
    exit_code,
    exp_pytest_failed,
    exp_msg,
    exp_kwargs,
    mock_pytest,
    mock_terminal_writer,
    mocker,
):
    mock_pytest.main.return_value = exit_code
    actual_pytest_failed = audit._run_pytest(mock_terminal_writer)

    assert exp_pytest_failed == actual_pytest_failed
    exp_calls = [
        mocker.call("Running tests for audit validation...\n"),
        mocker.call(exp_msg, **exp_kwargs),
    ]
    mock_terminal_writer.write.assert_has_calls(exp_calls)


def test_get_audit_steps(mock_terminal_writer, mocker, monkeypatch):
    mock_steps = [mocker.Mock() for i in range(3)]
    mock_get_plugins = mocker.Mock()
    mock_get_plugins.return_value = mock_steps
    monkeypatch.setattr(
        audit.plugin_utils, "load_plugins_by_namespace", mock_get_plugins
    )

    ret = audit._get_audit_steps("job/dir", "config", mock_terminal_writer)

    mock_get_plugins.assert_called_once_with("klio.plugins.audit")
    assert ret == [s.return_value for s in mock_steps]
    for s in mock_steps:
        s.assert_called_once_with("job/dir", "config", mock_terminal_writer)


@pytest.mark.parametrize("param_value", (None, "something"))
@pytest.mark.parametrize("res_parsing", (True, False))
def test_list_audit_steps(
    res_parsing, param_value, mock_terminal_writer, mocker, monkeypatch
):
    mock_click_context = mocker.Mock()
    mock_click_context.resilient_parsing = res_parsing
    mock_print_plugins = mocker.Mock()
    monkeypatch.setattr(
        audit.plugin_utils, "print_plugins", mock_print_plugins
    )

    audit.list_audit_steps(mock_click_context, None, param_value)

    if not param_value or res_parsing:
        mock_terminal_writer.sep.assert_not_called()
        mock_print_plugins.assert_not_called()
        mock_click_context.exit.assert_not_called()
    else:
        mock_terminal_writer.sep.assert_called_once_with(
            "=", "Installed audit steps"
        )
        mock_print_plugins.assert_called_once_with(
            "klio.plugins.audit", mock_terminal_writer
        )
        mock_click_context.exit.assert_called_once_with()


@pytest.mark.parametrize("steps_passed", (0, 1))
@pytest.mark.parametrize("steps_warned", (0, 1))
def test_audit_steps(
    steps_passed,
    steps_warned,
    mock_terminal_writer,
    mock_get_audit_steps,
    mock_run_pytest,
    mocker,
):

    mock_passed = []
    for i in range(steps_passed):
        name = "PassedStep{}".format(i + 1)
        mock_passed.append(
            mocker.Mock(AUDIT_STEP_NAME=name, warned=False, errored=False)
        )

    mock_warned = []
    for i in range(steps_warned):
        name = "WarnStep{}".format(i + 1)
        mock_warned.append(
            mocker.Mock(AUDIT_STEP_NAME=name, warned=True, errored=False)
        )

    mock_audit_steps = mock_passed + mock_warned
    mock_get_audit_steps.return_value = mock_audit_steps

    audit.audit("job/dir", "config")

    mock_get_audit_steps.assert_called_once_with(
        "job/dir", "config", mock_terminal_writer
    )
    for step in mock_audit_steps:
        step.before_tests.assert_called_once_with()

    mock_run_pytest.assert_called_once_with(mock_terminal_writer)

    # don't really care about the message strings, just that these funcs have
    # been called
    assert 2 == mock_terminal_writer.sep.call_count
    assert len(mock_passed) + 2 == mock_terminal_writer.write.call_count


def test_audit_steps_errored(
    mock_terminal_writer, mock_get_audit_steps, mock_run_pytest, mocker
):
    mock_errored = []
    for i in range(2):
        name = "WarnStep{}".format(i + 1)
        mock_errored.append(
            mocker.Mock(AUDIT_STEP_NAME=name, warned=False, errored=True)
        )

    mock_get_audit_steps.return_value = mock_errored

    with pytest.raises(SystemExit):
        audit.audit("job/dir", "config")

    mock_get_audit_steps.assert_called_once_with(
        "job/dir", "config", mock_terminal_writer
    )
    for step in mock_errored:
        step.before_tests.assert_called_once_with()

    mock_run_pytest.assert_called_once_with(mock_terminal_writer)

    # don't really care about the message strings, just that these funcs have
    # been called
    assert 3 == mock_terminal_writer.sep.call_count
    assert 1 == mock_terminal_writer.write.call_count


def test_audit_raises(mock_get_audit_steps, mock_terminal_writer, caplog):
    mock_get_audit_steps.side_effect = Exception("Blah")

    with pytest.raises(SystemExit) as exc_info:
        audit.audit("job/dir", "config")

    assert 1 == exc_info.value.code
    # don't really care about the message strings, just that these funcs have
    # been called
    assert 1 == mock_terminal_writer.write.call_count
    assert 1 == len(caplog.records)



# Copyright 2021 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from apache_beam.io.gcp import pubsub as beam_pubsub
from apache_beam.runners.direct import direct_runner
from apache_beam.runners.direct import transform_evaluator
from apache_beam.utils import timestamp as beam_timestamp
from google.api_core import exceptions as g_exceptions
from google.cloud import pubsub as g_pubsub

from klio.message import pubsub_message_manager as pmsg_mgr


class KlioPubSubReadEvaluator(transform_evaluator._PubSubReadEvaluator):
    """PubSubReadEvaluator for Klio's GkeDirectRunner.

    Behaves in the same way as _PubSubReadEvaluator, except for the fact
    that it acknowledges PubSub messages after they are done processing.
    """

    def __init__(self, *args, **kwargs):
        super(KlioPubSubReadEvaluator, self).__init__(*args, **kwargs)
        # Heads up: self._sub_name is from init'ing parent class
        self.sub_client = g_pubsub.SubscriberClient()
        self.message_manager = pmsg_mgr.MessageManager(self._sub_name)
        self.logger = logging.getLogger("klio.pubsub_read_evaluator")

    def _read_from_pubsub(self, timestamp_attribute):
        # Klio maintainer note: This code is the eact same logic in
        # _PubSubReadEvaluator._read_from_pubsub with the
        # following changes:
        # 1. Import statements that were originally inside this method
        #    was moved to the top of this module.
        # 2. Import statements adjusted to import module and not objects
        #    according to the google style guide.
        # 3. The functionalty we needed to override, which skips auto-acking
        #    consumed pubsub messages, and adds them to the MessageManager
        #    to handle deadline extension and acking once done.

        def _get_element(ack_id, message):
            parsed_message = beam_pubsub.PubsubMessage._from_message(message)
            if (
                timestamp_attribute
                and timestamp_attribute in parsed_message.attributes
            ):
                rfc3339_or_milli = parsed_message.attributes[
                    timestamp_attribute
                ]
                try:
                    timestamp = beam_timestamp.Timestamp(
                        micros=int(rfc3339_or_milli) * 1000
                    )
                except ValueError:
                    try:
                        timestamp = beam_timestamp.Timestamp.from_rfc3339(
                            rfc3339_or_milli
                        )
                    except ValueError as e:
                        raise ValueError("Bad timestamp value: %s" % e)
            else:
                timestamp = beam_timestamp.Timestamp(
                    message.publish_time.seconds,
                    message.publish_time.nanos // 1000,
                )

            self.message_manager.add(ack_id, parsed_message)

            return timestamp, parsed_message

        results = None
        try:
            response = self.sub_client.pull(
                self._sub_name, max_messages=1, return_immediately=True
            )
            results = [
                _get_element(rm.ack_id, rm.message)
                for rm in response.received_messages
            ]

        # only catching/ignoring this for now - if new exceptions raise, we'll
        # figure it out as they come on how to handle them
        except g_exceptions.DeadlineExceeded as e:
            # this seems mostly a benign error when there are 20+ seconds
            # between messages
            self.logger.debug(e)

        finally:
            self.sub_client.api.transport.channel.close()

        return results


class KlioTransformEvaluatorRegistry(
    transform_evaluator.TransformEvaluatorRegistry
):
    """TransformEvaluatorRegistry for Klio

    Makes DirectRunner's ReadFromPubSub use KlioPubSubReadEvaluator.
    """

    def __init__(self, *args, **kwargs):
        super(KlioTransformEvaluatorRegistry, self).__init__(*args, **kwargs)
        self._evaluators[
            direct_runner._DirectReadFromPubSub
        ] = KlioPubSubReadEvaluator
